# Copyright 2010 Jacob Kaplan-Moss

# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Server interface.
"""

import base64
import collections
from urllib import parse

from novaclient import api_versions
from novaclient import base
from novaclient import crypto
from novaclient import exceptions
from novaclient.i18n import _

_SENTINEL = object()

REBOOT_SOFT, REBOOT_HARD = 'SOFT', 'HARD'

CONSOLE_TYPE_ACTION_MAPPING = {
    'novnc': 'os-getVNCConsole',
    'xvpvnc': 'os-getVNCConsole',
    'spice-html5': 'os-getSPICEConsole',
    'rdp-html5': 'os-getRDPConsole',
    'serial': 'os-getSerialConsole'
}

CONSOLE_TYPE_PROTOCOL_MAPPING = {
    'novnc': 'vnc',
    'xvpvnc': 'vnc',
    'spice-html5': 'spice',
    'rdp-html5': 'rdp',
    'serial': 'serial',
    'webmks': 'mks'
}


class Server(base.Resource):
    HUMAN_ID = True

    def __repr__(self):
        return '<Server: %s>' % getattr(self, 'name', 'unknown-name')

    def delete(self):
        """
        Delete (i.e. shut down and delete the image) this server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.delete(self)

    @api_versions.wraps("2.0", "2.18")
    def update(self, name=None):
        """
        Update attributes of this server.

        :param name: Update the server's name.
        :returns: :class:`Server`
        """
        return self.manager.update(self, name=name)

    @api_versions.wraps("2.19", "2.89")
    def update(self, name=None, description=None):
        """
        Update attributes of this server.

        :param name: Update the server's name.
        :param description: Update the server's description.
        :returns: :class:`Server`
        """
        update_kwargs = {"name": name}
        if description is not None:
            update_kwargs["description"] = description
        return self.manager.update(self, **update_kwargs)

    @api_versions.wraps("2.90")
    def update(self, name=None, description=None, hostname=None):
        """
        Update attributes of this server.

        :param name: Update the server's name.
        :param description: Update the server's description.
        :param hostname: Update the server's hostname.
        :returns: :class:`Server`
        """
        update_kwargs = {"name": name}
        if description is not None:
            update_kwargs["description"] = description
        if hostname is not None:
            update_kwargs["hostname"] = hostname
        return self.manager.update(self, **update_kwargs)

    def get_console_output(self, length=None):
        """
        Get text console log output from Server.

        :param length: The number of lines you would like to retrieve (as int)
        """
        return self.manager.get_console_output(self, length)

    def get_vnc_console(self, console_type):
        """
        Get vnc console for a Server.

        :param console_type: Type of console ('novnc' or 'xvpvnc')
        """
        return self.manager.get_vnc_console(self, console_type)

    def get_spice_console(self, console_type):
        """
        Get spice console for a Server.

        :param console_type: Type of console ('spice-html5')
        """
        return self.manager.get_spice_console(self, console_type)

    def get_rdp_console(self, console_type):
        """
        Get rdp console for a Server.

        :param console_type: Type of console ('rdp-html5')
        """
        return self.manager.get_rdp_console(self, console_type)

    def get_serial_console(self, console_type):
        """
        Get serial console for a Server.

        :param console_type: Type of console ('serial')
        """
        return self.manager.get_serial_console(self, console_type)

    def get_mks_console(self):
        """
        Get mks console for a Server.

        """
        return self.manager.get_mks_console(self)

    def get_console_url(self, console_type):
        """
        Retrieve a console of a particular protocol and console_type

        :param console_type: Type of console
        """

        return self.manager.get_console_url(self, console_type)

    def get_password(self, private_key=None):
        """
        Get password for a Server.

        Returns the clear password of an instance if private_key is
        provided, returns the ciphered password otherwise.

        :param private_key: Path to private key file for decryption
                            (optional)
        """
        return self.manager.get_password(self, private_key)

    def clear_password(self):
        """
        Get password for a Server.

        """
        return self.manager.clear_password(self)

    def stop(self):
        """
        Stop -- Stop the running server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.stop(self)

    def force_delete(self):
        """
        Force delete -- Force delete a server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.force_delete(self)

    def restore(self):
        """
        Restore -- Restore a server in 'soft-deleted' state.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.restore(self)

    def start(self):
        """
        Start -- Start the paused server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.start(self)

    def pause(self):
        """
        Pause -- Pause the running server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.pause(self)

    def unpause(self):
        """
        Unpause -- Unpause the paused server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.unpause(self)

    @api_versions.wraps("2.0", "2.72")
    def lock(self):
        """
        Lock -- Lock the instance from certain operations.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.lock(self)

    @api_versions.wraps("2.73")
    def lock(self, reason=None):
        """
        Lock -- Lock the instance from certain operations.

        :param reason: (Optional) The lock reason.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.lock(self, reason=reason)

    def unlock(self):
        """
        Unlock -- Remove instance lock.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.unlock(self)

    def suspend(self):
        """
        Suspend -- Suspend the running server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.suspend(self)

    def resume(self):
        """
        Resume -- Resume the suspended server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.resume(self)

    def rescue(self, password=None, image=None):
        """
        Rescue -- Rescue the problematic server.

        :param password: The admin password to be set in the rescue instance.
        :param image: The :class:`Image` to rescue with.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.rescue(self, password, image)

    def unrescue(self):
        """
        Unrescue -- Unrescue the rescued server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.unrescue(self)

    def shelve(self):
        """
        Shelve -- Shelve the server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.shelve(self)

    def shelve_offload(self):
        """
        Shelve_offload -- Remove a shelved server from the compute node.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.shelve_offload(self)

    @api_versions.wraps("2.0", "2.76")
    def unshelve(self):
        """
        Unshelve -- Unshelve the server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.unshelve(self)

    @api_versions.wraps("2.77", "2.90")
    def unshelve(self, availability_zone=None):
        """
        Unshelve -- Unshelve the server.

        :param availability_zone: The specified availability zone name
                                  (Optional)
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.unshelve(self,
                                     availability_zone=availability_zone)

    @api_versions.wraps("2.91")
    def unshelve(self, availability_zone=object(), host=None):
        """
        Unshelve -- Unshelve the server.

        :param availability_zone: If specified the instance will be unshelved
                                  to the availability_zone.
                                  If None is passed the instance defined
                                  availability_zone is unpin and the instance
                                  will be scheduled to any availability_zone
                                  (free scheduling).
                                  If not specified the instance will be
                                  unshelved to either its defined
                                  availability_zone or any
                                  availability_zone (free scheduling).
        :param host: The specified host
                                  (Optional)
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if (
            availability_zone is None or isinstance(availability_zone, str)
        ) and host:
            return self.manager.unshelve(
                self, availability_zone=availability_zone, host=host)
        if availability_zone is None or isinstance(availability_zone, str):
            return self.manager.unshelve(
                self, availability_zone=availability_zone)
        if host:
            return self.manager.unshelve(self, host=host)
        return self.manager.unshelve(self)

    def diagnostics(self):
        """Diagnostics -- Retrieve server diagnostics."""
        return self.manager.diagnostics(self)

    @api_versions.wraps("2.78")
    def topology(self):
        """Retrieve server topology."""
        return self.manager.topology(self)

    @api_versions.wraps("2.0", "2.55")
    def migrate(self):
        """
        Migrate a server to a new host.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.migrate(self)

    @api_versions.wraps("2.56")
    def migrate(self, host=None):
        """
        Migrate a server to a new host.

        :param host: (Optional) The target host.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.migrate(self, host=host)

    def change_password(self, password):
        """
        Update the admin password for a server.

        :param password: string to set as the admin password on the server
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.change_password(self, password)

    def reboot(self, reboot_type=REBOOT_SOFT):
        """
        Reboot the server.

        :param reboot_type: either :data:`REBOOT_SOFT` for a software-level
                reboot, or `REBOOT_HARD` for a virtual power cycle hard reboot.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.reboot(self, reboot_type)

    # NOTE(stephenfin): It would be nice to make everything bar image a
    # kwarg-only argument but there are backwards-compatbility concerns
    def rebuild(
        self,
        image,
        password=None,
        preserve_ephemeral=False,
        *,
        disk_config=None,
        name=None,
        meta=None,
        files=None,
        description=_SENTINEL,
        key_name=_SENTINEL,
        userdata=_SENTINEL,
        trusted_image_certificates=_SENTINEL,
        hostname=_SENTINEL,
    ):
        """
        Rebuild -- shut down and then re-image -- this server.

        :param image: The :class:`Image` (or its ID) to re-image with.
        :param password: String to set as password on the rebuilt server.
        :param preserve_ephemeral: If True, request that any ephemeral device
            be preserved when rebuilding the instance. Defaults to False.
        :param disk_config: Partitioning mode to use on the rebuilt server.
            Valid values are 'AUTO' or 'MANUAL'
        :param name: Something to name the server.
        :param meta: A dict of arbitrary key/value metadata to store for this
            server. Both keys and values must be <=255 characters.
        :param files: A dict of files to overwrite on the server upon boot.
            Keys are file names (i.e. ``/etc/passwd``) and values are the file
            contents (either as a string or as a file-like object). A maximum
            of five entries is allowed, and each file must be 10k or less.
            (deprecated starting with microversion 2.57)
        :param description: Optional description of the server. If None is
            specified, the existing description will be unset.
            (starting from microversion 2.19)
        :param key_name: Optional key pair name for rebuild operation. If None
            is specified, the existing key will be unset.
            (starting from microversion 2.54)
        :param userdata: Optional user data to pass to be exposed by the
            metadata server; this can be a file type object as well or a
            string. If None is specified, the existing user_data is unset.
            (starting from microversion 2.57)
        :param trusted_image_certificates: A list of trusted certificate IDs
            or None to unset/reset the servers trusted image certificates
            (starting from microversion 2.63)
        :param hostname: Optional hostname to configure for the instance. If
            None is specified, the existing hostname will be unset.
            (starting from microversion 2.90)
        :returns: :class:`Server`
        """
        return self.manager.rebuild(
            self,
            image,
            password=password,
            disk_config=disk_config,
            preserve_ephemeral=preserve_ephemeral,
            name=name,
            meta=meta,
            files=files,
            description=description,
            key_name=key_name,
            userdata=userdata,
            trusted_image_certificates=trusted_image_certificates,
            hostname=hostname,
        )

    def resize(self, flavor, *, disk_config=None):
        """
        Resize the server's resources.

        :param flavor: The :class:`Flavor` (or its ID) to resize to.
        :param disk_config: Partitioning mode to use on the rebuilt server.
            Valid values are 'AUTO' or 'MANUAL'.
        :returns: An instance of novaclient.base.TupleWithMeta

        Until a resize event is confirmed with :meth:`confirm_resize`, the old
        server will be kept around and you'll be able to roll back to the old
        flavor quickly with :meth:`revert_resize`. All resizes are
        automatically confirmed after 24 hours by default.
        """
        return self.manager.resize(self, flavor, disk_config=disk_config)

    def create_image(self, image_name, metadata=None):
        """
        Create an image based on this server.

        :param image_name: The name to assign the newly create image.
        :param metadata: Metadata to assign to the image.
        """
        return self.manager.create_image(self, image_name, metadata)

    def backup(self, backup_name, backup_type, rotation):
        """
        Backup a server instance.

        :param backup_name: Name of the backup image
        :param backup_type: The backup type, like 'daily' or 'weekly'
        :param rotation: Int parameter representing how many backups to
                        keep around.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.backup(self, backup_name, backup_type, rotation)

    def confirm_resize(self):
        """
        Confirm that the resize worked, thus removing the original server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.confirm_resize(self)

    def revert_resize(self):
        """
        Revert a previous resize, switching back to the old server.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.revert_resize(self)

    @property
    def networks(self):
        """
        Generate a simplified list of addresses

        :returns: An OrderedDict, keyed by network name, and sorted by network
            name in ascending order.
        """
        networks = collections.OrderedDict()
        try:
            # Sort the keys by network name in natural (ascending) order.
            network_labels = sorted(self.addresses.keys())
            for network_label in network_labels:
                address_list = self.addresses[network_label]
                networks[network_label] = [a['addr'] for a in address_list]
            return networks
        except AttributeError:
            return {}

    @api_versions.wraps("2.0", "2.24")
    def live_migrate(self, host=None,
                     block_migration=False,
                     disk_over_commit=None):
        """
        Migrates a running instance to a new machine.

        :param host: destination host name.
        :param block_migration: if True, do block_migration, the default
                                value is False and None will be mapped to False
        :param disk_over_commit: if True, allow disk over commit, the default
                                 value is None which is mapped to False
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if block_migration is None:
            block_migration = False
        if disk_over_commit is None:
            disk_over_commit = False
        return self.manager.live_migrate(self, host,
                                         block_migration,
                                         disk_over_commit)

    @api_versions.wraps("2.25", "2.29")
    def live_migrate(self, host=None, block_migration=None):
        """
        Migrates a running instance to a new machine.

        :param host: destination host name.
        :param block_migration: if True, do block_migration, the default
                                value is None which is mapped to 'auto'.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if block_migration is None:
            block_migration = "auto"
        return self.manager.live_migrate(self, host, block_migration)

    @api_versions.wraps("2.30", "2.67")
    def live_migrate(self, host=None, block_migration=None, force=None):
        """
        Migrates a running instance to a new machine.

        :param host: destination host name.
        :param block_migration: if True, do block_migration, the default
                                value is None which is mapped to 'auto'.
        :param force: force to bypass the scheduler if host is provided.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if block_migration is None:
            block_migration = "auto"
        return self.manager.live_migrate(self, host, block_migration, force)

    @api_versions.wraps("2.68")
    def live_migrate(self, host=None, block_migration=None):
        """
        Migrates a running instance to a new machine.

        :param host: destination host name.
        :param block_migration: if True, do block_migration, the default
                                value is None which is mapped to 'auto'.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if block_migration is None:
            block_migration = "auto"
        return self.manager.live_migrate(self, host, block_migration)

    def reset_state(self, state='error'):
        """
        Reset the state of an instance to active or error.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.reset_state(self, state)

    def reset_network(self):
        """
        Reset network of an instance.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.reset_network(self)

    def add_security_group(self, security_group):
        """
        Add a security group to an instance.

        :param security_group: The name of security group to add
        :returns: An instance of novaclient.base.DictWithMeta
        """
        return self.manager.add_security_group(self, security_group)

    def remove_security_group(self, security_group):
        """
        Remove a security group from an instance.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.remove_security_group(self, security_group)

    def list_security_group(self):
        """
        List security group(s) of an instance.
        """
        return self.manager.list_security_group(self)

    @api_versions.wraps("2.0", "2.13")
    def evacuate(self, host=None, on_shared_storage=True, password=None):
        """
        Evacuate an instance from failed host to specified host.

        :param host: Name of the target host
        :param on_shared_storage: Specifies whether instance files located
                        on shared storage.
        :param password: string to set as admin password on the evacuated
                         server.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.evacuate(self, host, on_shared_storage, password)

    @api_versions.wraps("2.14", "2.28")
    def evacuate(self, host=None, password=None):
        """
        Evacuate an instance from failed host to specified host.

        :param host: Name of the target host
        :param password: string to set as admin password on the evacuated
                         server.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.evacuate(self, host, password)

    @api_versions.wraps("2.29", "2.67")
    def evacuate(self, host=None, password=None, force=None):
        """
        Evacuate an instance from failed host to specified host.

        :param host: Name of the target host
        :param password: string to set as admin password on the evacuated
                         server.
        :param force: forces to bypass the scheduler if host is provided.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.evacuate(self, host, password, force)

    @api_versions.wraps("2.68")
    def evacuate(self, host=None, password=None):
        """
        Evacuate an instance from failed host to specified host.

        :param host: Name of the target host
        :param password: string to set as admin password on the evacuated
                         server.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.evacuate(self, host, password)

    def interface_list(self):
        """
        List interfaces attached to an instance.
        """
        return self.manager.interface_list(self)

    @api_versions.wraps("2.0", "2.48")
    def interface_attach(self, port_id, net_id, fixed_ip):
        """
        Attach a network interface to an instance.
        """
        return self.manager.interface_attach(self, port_id, net_id, fixed_ip)

    @api_versions.wraps("2.49")
    def interface_attach(self, port_id, net_id, fixed_ip, tag=None):
        """
        Attach a network interface to an instance with an optional tag.
        """
        return self.manager.interface_attach(self, port_id, net_id, fixed_ip,
                                             tag)

    def interface_detach(self, port_id):
        """
        Detach a network interface from an instance.
        """
        return self.manager.interface_detach(self, port_id)

    def trigger_crash_dump(self):
        """Trigger crash dump in an instance"""
        return self.manager.trigger_crash_dump(self)

    @api_versions.wraps('2.26')
    def tag_list(self):
        """
        Get list of tags from an instance.
        """
        return self.manager.tag_list(self)

    @api_versions.wraps('2.26')
    def delete_tag(self, tag):
        """
        Remove single tag from an instance.
        """
        return self.manager.delete_tag(self, tag)

    @api_versions.wraps('2.26')
    def delete_all_tags(self):
        """
        Remove all tags from an instance.
        """
        return self.manager.delete_all_tags(self)

    @api_versions.wraps('2.26')
    def set_tags(self, tags):
        """
        Set list of tags to an instance.
        """
        return self.manager.set_tags(self, tags)

    @api_versions.wraps('2.26')
    def add_tag(self, tag):
        """
        Add single tag to an instance.
        """
        return self.manager.add_tag(self, tag)


class NetworkInterface(base.Resource):
    @property
    def id(self):
        return self.port_id

    def __repr__(self):
        return '<NetworkInterface: %s>' % self.id


class SecurityGroup(base.Resource):
    def __str__(self):
        return str(self.id)


class ServerManager(base.BootingManagerWithFind):
    resource_class = Server

    @staticmethod
    def transform_userdata(userdata):
        if hasattr(userdata, 'read'):
            userdata = userdata.read()

        # NOTE(melwitt): Text file data is converted to bytes prior to
        # base64 encoding. The utf-8 encoding will fail for binary files.
        try:
            userdata = userdata.encode("utf-8")
        except AttributeError:
            # In python 3, 'bytes' object has no attribute 'encode'
            pass

        return base64.b64encode(userdata).decode('utf-8')

    def _boot(
        self,
        response_key,
        name,
        image,
        flavor,
        meta=None,
        files=None,
        userdata=None,
        reservation_id=False,
        return_raw=False,
        min_count=None,
        max_count=None,
        security_groups=None,
        key_name=None,
        availability_zone=None,
        block_device_mapping=None,
        block_device_mapping_v2=None,
        nics=None,
        scheduler_hints=None,
        config_drive=None,
        admin_pass=None,
        disk_config=None,
        access_ip_v4=None,
        access_ip_v6=None,
        description=None,
        tags=None,
        trusted_image_certificates=None,
        host=None,
        hypervisor_hostname=None,
        hostname=None,
    ):
        """
        Create (boot) a new server.
        """
        body = {"server": {
            "name": name,
            "imageRef": str(base.getid(image)) if image else '',
            "flavorRef": str(base.getid(flavor)),
        }}
        if userdata:
            body["server"]["user_data"] = self.transform_userdata(userdata)
        if meta:
            body["server"]["metadata"] = meta
        if reservation_id:
            body["server"]["return_reservation_id"] = reservation_id
            return_raw = True
        if key_name:
            body["server"]["key_name"] = key_name
        if scheduler_hints:
            body['os:scheduler_hints'] = scheduler_hints
        if config_drive:
            body["server"]["config_drive"] = config_drive
        if admin_pass:
            body["server"]["adminPass"] = admin_pass
        if not min_count:
            min_count = 1
        if not max_count:
            max_count = min_count
        body["server"]["min_count"] = min_count
        body["server"]["max_count"] = max_count

        if security_groups:
            body["server"]["security_groups"] = [{'name': sg}
                                                 for sg in security_groups]

        # Files are a slight bit tricky. They're passed in a "personality"
        # list to the POST. Each item is a dict giving a file name and the
        # base64-encoded contents of the file. We want to allow passing
        # either an open file *or* some contents as files here.
        if files:
            personality = body['server']['personality'] = []
            for filepath, file_or_string in sorted(files.items(),
                                                   key=lambda x: x[0]):
                if hasattr(file_or_string, 'read'):
                    data = file_or_string.read()
                else:
                    data = file_or_string

                if isinstance(data, str):
                    data = data.encode('utf-8')
                cont = base64.b64encode(data).decode('utf-8')
                personality.append({
                    'path': filepath,
                    'contents': cont,
                })

        if availability_zone:
            body["server"]["availability_zone"] = availability_zone

        # Block device mappings are passed as a list of dictionaries
        # in the create API
        if block_device_mapping:
            body['server']['block_device_mapping'] = \
                self._parse_block_device_mapping(block_device_mapping)
        elif block_device_mapping_v2:
            # Following logic can't be removed because it will leaves
            # a valid boot with both --image and --block-device
            # failed , see bug 1433609 for more info
            if image:
                bdm_dict = {'uuid': base.getid(image), 'source_type': 'image',
                            'destination_type': 'local', 'boot_index': 0,
                            'delete_on_termination': True}
                block_device_mapping_v2.insert(0, bdm_dict)
            body['server']['block_device_mapping_v2'] = block_device_mapping_v2

        if nics is not None:
            # With microversion 2.37+ nics can be an enum of 'auto' or 'none'
            # or a list of dicts.
            if isinstance(nics, str):
                all_net_data = nics
            else:
                # NOTE(tr3buchet): nics can be an empty list
                all_net_data = []
                for nic_info in nics:
                    net_data = {}
                    # if value is empty string, do not send value in body
                    if nic_info.get('net-id'):
                        net_data['uuid'] = nic_info['net-id']
                    if (nic_info.get('v4-fixed-ip') and
                            nic_info.get('v6-fixed-ip')):
                        raise base.exceptions.CommandError(_(
                            "Only one of 'v4-fixed-ip' and 'v6-fixed-ip' "
                            "may be provided."))
                    elif nic_info.get('v4-fixed-ip'):
                        net_data['fixed_ip'] = nic_info['v4-fixed-ip']
                    elif nic_info.get('v6-fixed-ip'):
                        net_data['fixed_ip'] = nic_info['v6-fixed-ip']
                    if nic_info.get('port-id'):
                        net_data['port'] = nic_info['port-id']
                    if nic_info.get('tag'):
                        net_data['tag'] = nic_info['tag']
                    all_net_data.append(net_data)
            body['server']['networks'] = all_net_data

        if disk_config is not None:
            body['server']['OS-DCF:diskConfig'] = disk_config

        if access_ip_v4 is not None:
            body['server']['accessIPv4'] = access_ip_v4

        if access_ip_v6 is not None:
            body['server']['accessIPv6'] = access_ip_v6

        if description:
            body['server']['description'] = description

        if tags:
            body['server']['tags'] = tags

        if trusted_image_certificates:
            body['server']['trusted_image_certificates'] = (
                trusted_image_certificates)

        if host:
            body['server']['host'] = host

        if hypervisor_hostname:
            body['server']['hypervisor_hostname'] = hypervisor_hostname

        if hostname:
            body['server']['hostname'] = hostname

        return self._create(
            '/servers', body, response_key, return_raw=return_raw,
        )

    def get(self, server):
        """
        Get a server.

        :param server: ID of the :class:`Server` to get.
        :rtype: :class:`Server`
        """
        return self._get("/servers/%s" % base.getid(server), "server")

    def list(self, detailed=True, search_opts=None, marker=None, limit=None,
             sort_keys=None, sort_dirs=None):
        """
        Get a list of servers.

        :param detailed: Whether to return detailed server info (optional).
        :param search_opts: Search options to filter out servers which don't
            match the search_opts (optional). The search opts format is a
            dictionary of key / value pairs that will be appended to the query
            string.  For a complete list of keys see:
            https://docs.openstack.org/api-ref/compute/#list-servers
        :param marker: Begin returning servers that appear later in the server
                       list than that represented by this server id (optional).
        :param limit: Maximum number of servers to return (optional).
                      Note the API server has a configurable default limit.
                      If no limit is specified here or limit is larger than
                      default, the default limit will be used.
                      If limit == -1, all servers will be returned.
        :param sort_keys: List of sort keys
        :param sort_dirs: List of sort directions

        :rtype: list of :class:`Server`

        Examples:

        client.servers.list() - returns detailed list of servers

        client.servers.list(search_opts={'status': 'ERROR'}) -
        returns list of servers in error state.

        client.servers.list(limit=10) - returns only 10 servers

        """
        if search_opts is None:
            search_opts = {}

        qparams = {}
        # In microversion 2.73 we added ``locked`` filtering option
        # for listing server details.
        if ('locked' in search_opts and
                self.api_version < api_versions.APIVersion('2.73')):
            raise exceptions.UnsupportedAttribute("locked", "2.73")
        for opt, val in search_opts.items():
            # support locked=False from 2.73 microversion
            if val or (opt == 'locked' and val is False):
                if isinstance(val, str):
                    val = val.encode('utf-8')
                qparams[opt] = val
            # NOTE(gibi): The False value won't actually do anything until we
            # fix bug 1871409 and clean up the API inconsistency, but we do it
            # in preparation for that (hopefully backportable) fix
            if opt == 'config_drive' and val is not None:
                qparams[opt] = str(val)

        detail = ""
        if detailed:
            detail = "/detail"

        result = base.ListWithMeta([], None)
        while True:
            if marker:
                qparams['marker'] = marker

            if limit and limit != -1:
                qparams['limit'] = limit

            # Transform the dict to a sequence of two-element tuples in fixed
            # order, then the encoded string will be consistent in Python 2&3.
            if qparams or sort_keys or sort_dirs:
                # sort keys and directions are unique since the same parameter
                # key is repeated for each associated value
                # (ie, &sort_key=key1&sort_key=key2&sort_key=key3)
                items = list(qparams.items())
                if sort_keys:
                    items.extend(('sort_key', sort_key)
                                 for sort_key in sort_keys)
                if sort_dirs:
                    items.extend(('sort_dir', sort_dir)
                                 for sort_dir in sort_dirs)
                new_qparams = sorted(items, key=lambda x: x[0])
                query_string = "?%s" % parse.urlencode(new_qparams)
            else:
                query_string = ""

            servers = self._list("/servers%s%s" % (detail, query_string),
                                 "servers")
            result.extend(servers)
            result.append_request_ids(servers.request_ids)

            if not servers or limit != -1:
                break
            marker = result[-1].id
        return result

    def get_vnc_console(self, server, console_type):
        """
        Get a vnc console for an instance

        :param server: The :class:`Server` (or its ID) to get console for.
        :param console_type: Type of vnc console to get ('novnc' or 'xvpvnc')
        :returns: An instance of novaclient.base.DictWithMeta
        """

        return self.get_console_url(server, console_type)

    def get_spice_console(self, server, console_type):
        """
        Get a spice console for an instance

        :param server: The :class:`Server` (or its ID) to get console for.
        :param console_type: Type of spice console to get ('spice-html5')
        :returns: An instance of novaclient.base.DictWithMeta
        """

        return self.get_console_url(server, console_type)

    def get_rdp_console(self, server, console_type):
        """
        Get a rdp console for an instance

        :param server: The :class:`Server` (or its ID) to get console for.
        :param console_type: Type of rdp console to get ('rdp-html5')
        :returns: An instance of novaclient.base.DictWithMeta
        """

        return self.get_console_url(server, console_type)

    def get_serial_console(self, server, console_type):
        """
        Get a serial console for an instance

        :param server: The :class:`Server` (or its ID) to get console for.
        :param console_type: Type of serial console to get ('serial')
        :returns: An instance of novaclient.base.DictWithMeta
        """

        return self.get_console_url(server, console_type)

    def _get_protocol(self, console_type):
        protocol = CONSOLE_TYPE_PROTOCOL_MAPPING.get(console_type)
        if not protocol:
            raise exceptions.UnsupportedConsoleType(console_type)

        return protocol

    @api_versions.wraps('2.0', '2.5')
    def get_console_url(self, server, console_type):
        """
        Retrieve a console url of a server.

        :param server: server to get console url for
        :param console_type: type can be novnc, xvpvnc, spice-html5,
                             rdp-html5 and serial.
        """

        action = CONSOLE_TYPE_ACTION_MAPPING.get(console_type)
        if not action:
            raise exceptions.UnsupportedConsoleType(console_type)
        return self._action(action, server, {'type': console_type})

    @api_versions.wraps('2.8')
    def get_mks_console(self, server):
        """
        Get a mks console for an instance

        :param server: The :class:`Server` (or its ID) to get console for.
        :returns: An instance of novaclient.base.DictWithMeta
        """

        return self.get_console_url(server, 'webmks')

    @api_versions.wraps('2.6')
    def get_console_url(self, server, console_type):
        """
        Retrieve a console url of a server.

        :param server: server to get console url for
        :param console_type: type can be novnc/xvpvnc for protocol vnc;
                             spice-html5 for protocol spice; rdp-html5 for
                             protocol rdp; serial for protocol serial.
                             webmks for protocol mks (since version 2.8).
        """

        if self.api_version < api_versions.APIVersion('2.8'):
            if console_type == 'webmks':
                raise exceptions.UnsupportedConsoleType(console_type)

        protocol = self._get_protocol(console_type)
        body = {'remote_console': {'protocol': protocol,
                                   'type': console_type}}
        url = '/servers/%s/remote-consoles' % base.getid(server)
        resp, body = self.api.client.post(url, body=body)
        return self.convert_into_with_meta(body, resp)

    def get_password(self, server, private_key=None):
        """
        Get admin password of an instance

        Returns the admin password of an instance in the clear if private_key
        is provided, returns the ciphered password otherwise.

        Requires that openssl is installed and in the path

        :param server: The :class:`Server` (or its ID) for which the admin
                       password is to be returned
        :param private_key: The private key to decrypt password
                            (optional)
        :returns: An instance of novaclient.base.StrWithMeta or
                  novaclient.base.BytesWithMeta or
                  novaclient.base.UnicodeWithMeta
        """

        resp, body = self.api.client.get("/servers/%s/os-server-password"
                                         % base.getid(server))
        ciphered_pw = body.get('password', '') if body else ''
        if private_key and ciphered_pw:
            try:
                ciphered_pw = crypto.decrypt_password(private_key, ciphered_pw)
            except Exception as exc:
                ciphered_pw = '%sFailed to decrypt:\n%s' % (exc, ciphered_pw)
        return self.convert_into_with_meta(ciphered_pw, resp)

    def clear_password(self, server):
        """
        Clear the admin password of an instance

        Remove the admin password for an instance from the metadata server.

        :param server: The :class:`Server` (or its ID) for which the admin
                       password is to be cleared
        """

        return self._delete("/servers/%s/os-server-password"
                            % base.getid(server))

    def stop(self, server):
        """
        Stop the server.

        :param server: The :class:`Server` (or its ID) to stop
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        resp, body = self._action_return_resp_and_body('os-stop', server, None)
        return base.TupleWithMeta((resp, body), resp)

    def force_delete(self, server):
        """
        Force delete the server.

        :param server: The :class:`Server` (or its ID) to force delete
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        resp, body = self._action_return_resp_and_body('forceDelete', server,
                                                       None)
        return base.TupleWithMeta((resp, body), resp)

    def restore(self, server):
        """
        Restore soft-deleted server.

        :param server: The :class:`Server` (or its ID) to restore
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        resp, body = self._action_return_resp_and_body('restore', server, None)
        return base.TupleWithMeta((resp, body), resp)

    def start(self, server):
        """
        Start the server.

        :param server: The :class:`Server` (or its ID) to start
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('os-start', server, None)

    def pause(self, server):
        """
        Pause the server.

        :param server: The :class:`Server` (or its ID) to pause
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('pause', server, None)

    def unpause(self, server):
        """
        Unpause the server.

        :param server: The :class:`Server` (or its ID) to unpause
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('unpause', server, None)

    @api_versions.wraps("2.0", "2.72")
    def lock(self, server):
        """
        Lock the server.

        :param server: The :class:`Server` (or its ID) to lock
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('lock', server, None)

    @api_versions.wraps("2.73")
    def lock(self, server, reason=None):
        """
        Lock the server.

        :param server: The :class:`Server` (or its ID) to lock
        :param reason: (Optional) The lock reason.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        info = None

        if reason:
            info = {'locked_reason': reason}

        return self._action('lock', server, info)

    def unlock(self, server):
        """
        Unlock the server.

        :param server: The :class:`Server` (or its ID) to unlock
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('unlock', server, None)

    def suspend(self, server):
        """
        Suspend the server.

        :param server: The :class:`Server` (or its ID) to suspend
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('suspend', server, None)

    def resume(self, server):
        """
        Resume the server.

        :param server: The :class:`Server` (or its ID) to resume
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('resume', server, None)

    def rescue(self, server, password=None, image=None):
        """
        Rescue the server.

        :param server: The :class:`Server` to rescue.
        :param password: The admin password to be set in the rescue instance.
        :param image: The :class:`Image` to rescue with.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        info = {}
        if password:
            info['adminPass'] = password
        if image:
            info['rescue_image_ref'] = base.getid(image)
        resp, body = self._action_return_resp_and_body('rescue', server,
                                                       info or None)
        return base.TupleWithMeta((resp, body), resp)

    def unrescue(self, server):
        """
        Unrescue the server.

        :param server: The :class:`Server` (or its ID) to unrescue
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('unrescue', server, None)

    def shelve(self, server):
        """
        Shelve the server.

        :param server: The :class:`Server` (or its ID) to shelve
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('shelve', server, None)

    def shelve_offload(self, server):
        """
        Remove a shelved instance from the compute node.

        :param server: The :class:`Server` (or its ID) to shelve offload
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('shelveOffload', server, None)

    @api_versions.wraps("2.0", "2.76")
    def unshelve(self, server):
        """
        Unshelve the server.

        :param server: The :class:`Server` (or its ID) to unshelve
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('unshelve', server, None)

    @api_versions.wraps("2.77", "2.90")
    def unshelve(self, server, availability_zone=None):
        """
        Unshelve the server.

        :param server: The :class:`Server` (or its ID) to unshelve
        :param availability_zone: The specified availability zone name
                                  (Optional)
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        info = None
        if availability_zone:
            info = {'availability_zone': availability_zone}
        return self._action('unshelve', server, info)

    @api_versions.wraps("2.91")
    def unshelve(self, server, availability_zone=object(), host=None):
        """
        Unshelve the server.

        :param availability_zone: If specified the instance will be unshelved
                                  to the availability_zone.
                                  If None is passed the instance defined
                                  availability_zone is unpin and the instance
                                  will be scheduled to any availability_zone
                                  (free scheduling).
                                  If not specified the instance will be
                                  unshelved to either its defined
                                  availability_zone or any
                                  availability_zone (free scheduling).
        :param host: The specified host
                                  (Optional)
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        info = None

        if availability_zone is None or isinstance(availability_zone, str):
            info = {'availability_zone': availability_zone}
        if host:
            if info:
                info['host'] = host
            else:
                info = {'host': host}
        return self._action('unshelve', server, info)

    def ips(self, server):
        """
        Return IP Addresses associated with the server.

        Often a cheaper call then getting all the details for a server.

        :param server: The :class:`Server` (or its ID) for which
                       the IP addresses are to be returned
        :returns: An instance of novaclient.base.DictWithMeta
        """
        resp, body = self.api.client.get("/servers/%s/ips" %
                                         base.getid(server))
        return base.DictWithMeta(body['addresses'], resp)

    def diagnostics(self, server):
        """
        Retrieve server diagnostics.

        :param server: The :class:`Server` (or its ID) for which
                       diagnostics to be returned
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        resp, body = self.api.client.get("/servers/%s/diagnostics" %
                                         base.getid(server))
        return base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps("2.78")
    def topology(self, server):
        """
        Retrieve server topology.

        :param server: The :class:`Server` (or its ID) for which
                       topology to be returned
        :returns: An instance of novaclient.base.DictWithMeta
        """
        resp, body = self.api.client.get("/servers/%s/topology" %
                                         base.getid(server))
        return base.DictWithMeta(body, resp)

    def _validate_create_nics(self, nics):
        # nics are required with microversion 2.37+ and can be a string or list
        if self.api_version > api_versions.APIVersion('2.36'):
            if not nics:
                raise ValueError('nics are required after microversion 2.36')
        elif nics and not isinstance(nics, (list, tuple)):
            raise ValueError('nics must be a list or a tuple, not %s' %
                             type(nics))

    # NOTE(stephenfin): It would be nice to make everything bar name, image and
    # flavor a kwarg-only argument but there are backwards-compatbility
    # concerns
    def create(
        self,
        name,
        image,
        flavor,
        meta=None,
        files=None,
        reservation_id=False,
        min_count=None,
        max_count=None,
        security_groups=None,
        userdata=None,
        key_name=None,
        availability_zone=None,
        block_device_mapping=None,
        block_device_mapping_v2=None,
        nics=None,
        scheduler_hints=None,
        config_drive=None,
        disk_config=None,
        admin_pass=None,
        access_ip_v4=None,
        access_ip_v6=None,
        description=None,
        tags=None,
        trusted_image_certificates=None,
        host=None,
        hypervisor_hostname=None,
        hostname=None,
    ):
        """
        Create (boot) a new server.

        In order to create a server with pre-existing ports that contain a
        ``resource_request`` value, such as for guaranteed minimum bandwidth
        quality of service support, microversion ``2.72`` is required.

        :param name: Something to name the server.
        :param image: The :class:`Image` to boot with.
        :param flavor: The :class:`Flavor` to boot onto.
        :param meta: A dict of arbitrary key/value metadata to store for this
            server. Both keys and values must be <=255 characters.
        :param files: A dict of files to overwrite on the server upon boot.
            Keys are file names (i.e. ``/etc/passwd``) and values
            are the file contents (either as a string or as a
            file-like object). A maximum of five entries is allowed,
            and each file must be 10k or less.
            (deprecated starting with microversion 2.57)
        :param reservation_id: Return a reservation_id for the set of
            servers being requested, boolean.
        :param min_count: The minimum number of servers to launch.
        :param max_count: The maximum number of servers to launch.
        :param security_groups: A list of security group names
        :param userdata: User data to pass to be exposed by the metadata
            server this can be a file type object as well or a string.
        :param key_name: Name of previously created keypair to inject into the
            instance.
        :param availability_zone: Name of the availability zone for instance
            placement.
        :param block_device_mapping: A dict of block device mappings for this
            server.
        :param block_device_mapping_v2: A list of block device mappings (dicts)
            for this server.
        :param nics: An ordered list of nics (dicts) to be added to this
            server, with information about connected networks, fixed IPs, port
            etc. Beginning in microversion 2.37 this field is required and also
            supports a single string value of 'auto' or 'none'. The 'auto'
            value means the Compute service will automatically allocate a
            network for the project if one is not available. This is the same
            behavior as not passing anything for nics before microversion 2.37.
            The 'none' value tells the Compute service to not allocate any
            networking for the server.
        :param scheduler_hints: Arbitrary key-value pairs specified by the
            client to help boot an instance.
        :param config_drive: A boolean value to enable config drive.
        :param disk_config: Control how the disk is partitioned when the server
            is created. Possible values are 'AUTO' or 'MANUAL'.
        :param admin_pass: Add a user supplied admin password.
        :param access_ip_v4: Add alternative access IP (v4)
        :param access_ip_v6: Add alternative access IP (v6)
        :param description: Optional description of the server
            (allowed since microversion 2.19)
        :param tags: A list of arbitrary strings to be added to the server as
            tags
            (allowed since microversion 2.52)
        :param trusted_image_certificates: A list of trusted certificate IDs
            (allowed since microversion 2.63)
        :param host: Requested host to create servers
            (allowed since microversion 2.74)
        :param hypervisor_hostname: Requested hypervisor hostname to create
            servers
            (allowed since microversion 2.74)
        :param hostname: Requested hostname of server
            (allowed since microversion 2.90)
        """
        if not min_count:
            min_count = 1
        if not max_count:
            max_count = min_count
        if min_count > max_count:
            min_count = max_count

        boot_args = [name, image, flavor]

        if description and self.api_version < api_versions.APIVersion("2.19"):
            raise exceptions.UnsupportedAttribute("description", "2.19")

        self._validate_create_nics(nics)

        tags_microversion = api_versions.APIVersion("2.32")
        if self.api_version < tags_microversion:
            if nics:
                for nic_info in nics:
                    if nic_info.get("tag"):
                        raise ValueError("Setting interface tags is "
                                         "unsupported before microversion "
                                         "2.32")

            if block_device_mapping_v2:
                for bdm in block_device_mapping_v2:
                    if bdm.get("tag"):
                        raise ValueError("Setting block device tags is "
                                         "unsupported before microversion "
                                         "2.32")

        if tags and self.api_version < api_versions.APIVersion("2.52"):
            raise exceptions.UnsupportedAttribute("tags", "2.52")

        personality_files_deprecation = api_versions.APIVersion('2.57')
        if files and self.api_version >= personality_files_deprecation:
            raise exceptions.UnsupportedAttribute('files', '2.0', '2.56')

        trusted_certs_microversion = api_versions.APIVersion("2.63")
        if (trusted_image_certificates and
                self.api_version < trusted_certs_microversion):
            raise exceptions.UnsupportedAttribute("trusted_image_certificates",
                                                  "2.63")

        if (block_device_mapping_v2 and
                self.api_version < api_versions.APIVersion('2.67')):
            for bdm in block_device_mapping_v2:
                if bdm.get('volume_type'):
                    raise ValueError(
                        "Block device volume_type is not supported before "
                        "microversion 2.67")

        host_microversion = api_versions.APIVersion("2.74")
        if host and self.api_version < host_microversion:
            raise exceptions.UnsupportedAttribute("host", "2.74")
        hypervisor_hostname_microversion = api_versions.APIVersion("2.74")
        if (hypervisor_hostname and
                self.api_version < hypervisor_hostname_microversion):
            raise exceptions.UnsupportedAttribute(
                "hypervisor_hostname", "2.74")

        hostname_microversion = api_versions.APIVersion("2.90")
        if hostname and self.api_version < hostname_microversion:
            raise exceptions.UnsupportedAttribute("hostname", "2.90")

        boot_kwargs = dict(
            meta=meta, files=files, userdata=userdata,
            reservation_id=reservation_id, min_count=min_count,
            max_count=max_count, security_groups=security_groups,
            key_name=key_name, availability_zone=availability_zone,
            scheduler_hints=scheduler_hints, config_drive=config_drive,
            disk_config=disk_config, admin_pass=admin_pass,
            access_ip_v4=access_ip_v4, access_ip_v6=access_ip_v6,
            description=description, tags=tags,
            trusted_image_certificates=trusted_image_certificates,
            host=host, hypervisor_hostname=hypervisor_hostname,
            hostname=hostname)

        if block_device_mapping:
            boot_kwargs['block_device_mapping'] = block_device_mapping
        elif block_device_mapping_v2:
            boot_kwargs['block_device_mapping_v2'] = block_device_mapping_v2

        if nics:
            boot_kwargs['nics'] = nics

        response_key = "server" if not reservation_id else "reservation_id"
        return self._boot(response_key, *boot_args, **boot_kwargs)

    @api_versions.wraps("2.0", "2.18")
    def update(self, server, name=None):
        """
        Update attributes of a server.

        :param server: The :class:`Server` (or its ID) to update.
        :param name: Update the server's name.
        :returns: :class:`Server`
        """
        if name is None:
            return

        body = {
            "server": {
                "name": name,
            },
        }

        return self._update("/servers/%s" % base.getid(server), body, "server")

    @api_versions.wraps("2.19", "2.89")
    def update(self, server, name=None, description=None):
        """
        Update attributes of a server.

        :param server: The :class:`Server` (or its ID) to update.
        :param name: Update the server's name.
        :param description: Update the server's description. If it equals to
            empty string(i.g. ""), the server description will be removed.
        :returns: :class:`Server`
        """
        if name is None and description is None:
            return

        body = {"server": {}}
        if name:
            body["server"]["name"] = name
        if description == "":
            body["server"]["description"] = None
        elif description:
            body["server"]["description"] = description

        return self._update("/servers/%s" % base.getid(server), body, "server")

    @api_versions.wraps("2.90")
    def update(self, server, name=None, description=None, hostname=None):
        """
        Update attributes of a server.

        :param server: The :class:`Server` (or its ID) to update.
        :param name: Update the server's name.
        :param description: Update the server's description. If it equals to
            empty string(i.g. ""), the server description will be removed.
        :param hostname: Update the server's hostname as recorded by the
            metadata service. Note that a separate utility running on the
            guest will be necessary to reflect these changes in the guest
            itself.
        :returns: :class:`Server`
        """
        if name is None and description is None and hostname is None:
            return

        body = {"server": {}}
        if name:
            body["server"]["name"] = name
        if description == "":
            body["server"]["description"] = None
        elif description:
            body["server"]["description"] = description
        if hostname:
            body["server"]["hostname"] = hostname

        return self._update("/servers/%s" % base.getid(server), body, "server")

    def change_password(self, server, password):
        """
        Update the password for a server.

        :param server: The :class:`Server` (or its ID) for which the admin
                       password is to be changed
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action("changePassword", server, {"adminPass": password})

    def delete(self, server):
        """
        Delete (i.e. shut down and delete the image) this server.

        :param server: The :class:`Server` (or its ID) to delete
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._delete("/servers/%s" % base.getid(server))

    def reboot(self, server, reboot_type=REBOOT_SOFT):
        """
        Reboot a server.

        :param server: The :class:`Server` (or its ID) to share onto.
        :param reboot_type: either :data:`REBOOT_SOFT` for a software-level
                reboot, or `REBOOT_HARD` for a virtual power cycle hard reboot.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('reboot', server, {'type': reboot_type})

    # NOTE(stephenfin): It would be nice to make everything bar server and
    # image a kwarg-only argument but there are backwards-compatbility concerns
    def rebuild(
        self,
        server,
        image,
        password=None,
        disk_config=None,
        preserve_ephemeral=False,
        name=None,
        meta=None,
        files=None,
        *,
        description=_SENTINEL,
        key_name=_SENTINEL,
        userdata=_SENTINEL,
        trusted_image_certificates=_SENTINEL,
        hostname=_SENTINEL,
    ):
        """
        Rebuild -- shut down and then re-image -- a server.

        :param server: The :class:`Server` (or its ID) to share onto.
        :param image: The :class:`Image` (or its ID) to re-image with.
        :param password: String to set as password on the rebuilt server.
        :param disk_config: Partitioning mode to use on the rebuilt server.
            Valid values are 'AUTO' or 'MANUAL'
        :param preserve_ephemeral: If True, request that any ephemeral device
            be preserved when rebuilding the instance. Defaults to False.
        :param name: Something to name the server.
        :param meta: A dict of arbitrary key/value metadata to store for this
            server. Both keys and values must be <=255 characters.
        :param files: A dict of files to overwrite on the server upon boot.
            Keys are file names (i.e. ``/etc/passwd``) and values are the file
            contents (either as a string or as a file-like object). A maximum
            of five entries is allowed, and each file must be 10k or less.
            (deprecated starting with microversion 2.57)
        :param description: Optional description of the server. If None is
            specified, the existing description will be unset.
            (starting from microversion 2.19)
        :param key_name: Optional key pair name for rebuild operation. If None
            is specified, the existing key will be unset.
            (starting from microversion 2.54)
        :param userdata: Optional user data to pass to be exposed by the
            metadata server; this can be a file type object as well or a
            string. If None is specified, the existing user_data is unset.
            (starting from microversion 2.57)
        :param trusted_image_certificates: A list of trusted certificate IDs
            or None to unset/reset the servers trusted image certificates
            (starting from microversion 2.63)
        :param hostname: Optional hostname to configure for the instance. If
            None is specified, the existing hostname will be unset.
            (starting from microversion 2.90)
        :returns: :class:`Server`
        """
        # Microversion 2.19 adds the optional 'description' parameter
        if (
            description is not _SENTINEL and
            self.api_version < api_versions.APIVersion('2.19')
        ):
            raise exceptions.UnsupportedAttribute('description', '2.19')

        # Microversion 2.54 adds the optional 'key_name' parameter
        if (
            key_name is not _SENTINEL and
            self.api_version < api_versions.APIVersion('2.54')
        ):
            raise exceptions.UnsupportedAttribute('key_name', '2.54')

        # Microversion 2.57 deprecates personality files and adds support
        # for user_data.
        if files and self.api_version >= api_versions.APIVersion('2.57'):
            raise exceptions.UnsupportedAttribute('files', '2.0', '2.56')

        if (
            userdata is not _SENTINEL and
            self.api_version < api_versions.APIVersion('2.57')
        ):
            raise exceptions.UnsupportedAttribute('userdata', '2.57')

        # Microversion 2.63 adds trusted image certificate support
        if (
            trusted_image_certificates is not _SENTINEL and
            self.api_version < api_versions.APIVersion('2.63')
        ):
            raise exceptions.UnsupportedAttribute(
                'trusted_image_certificates', '2.63')

        # Microversion 2.90 adds the optional 'hostname' parameter
        if (
            hostname is not _SENTINEL and
            self.api_version < api_versions.APIVersion('2.90')
        ):
            raise exceptions.UnsupportedAttribute('hostname', '2.90')

        body = {'imageRef': base.getid(image)}

        if password is not None:
            body['adminPass'] = password

        if disk_config is not None:
            body['OS-DCF:diskConfig'] = disk_config

        if preserve_ephemeral is not False:
            body['preserve_ephemeral'] = True

        if name is not None:
            body['name'] = name

        if description is not _SENTINEL:
            body["description"] = description

        if key_name is not _SENTINEL:
            body['key_name'] = key_name

        if trusted_image_certificates is not _SENTINEL:
            body["trusted_image_certificates"] = trusted_image_certificates

        if hostname is not _SENTINEL:
            body["hostname"] = hostname

        if meta:
            body['metadata'] = meta

        if files:
            personality = body['personality'] = []
            for filepath, file_or_string in sorted(
                files.items(), key=lambda x: x[0],
            ):
                if hasattr(file_or_string, 'read'):
                    data = file_or_string.read()
                else:
                    data = file_or_string

                cont = base64.b64encode(data.encode('utf-8')).decode('utf-8')
                personality.append({
                    'path': filepath,
                    'contents': cont,
                })

        if userdata is not _SENTINEL:
            # If userdata is specified but None, it means unset the existing
            # user_data on the instance.
            userdata = userdata
            body['user_data'] = (userdata if userdata is None else
                                 self.transform_userdata(userdata))

        resp, body = self._action_return_resp_and_body(
            'rebuild', server, body,
        )
        return Server(self, body['server'], resp=resp)

    @api_versions.wraps("2.0", "2.55")
    def migrate(self, server):
        """
        Migrate a server to a new host.

        :param server: The :class:`Server` (or its ID).
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('migrate', server)

    @api_versions.wraps("2.56")
    def migrate(self, server, host=None):
        """
        Migrate a server to a new host.

        :param server: The :class:`Server` (or its ID).
        :param host: (Optional) The target host.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        info = {}

        if host:
            info['host'] = host

        return self._action('migrate', server, info)

    # NOTE(stephenfin): It would be nice to make disk_config a kwarg-only
    # argument but there are backwards-compatbility concerns
    def resize(self, server, flavor, disk_config=None):
        """
        Resize a server's resources.

        :param server: The :class:`Server` (or its ID) to share onto.
        :param flavor: The :class:`Flavor` (or its ID) to resize to.
        :param disk_config: Partitioning mode to use on the rebuilt server.
            Valid values are 'AUTO' or 'MANUAL'.
        :returns: An instance of novaclient.base.TupleWithMeta

        Until a resize event is confirmed with :meth:`confirm_resize`, the old
        server will be kept around and you'll be able to roll back to the old
        flavor quickly with :meth:`revert_resize`. All resizes are
        automatically confirmed after 24 hours by default.
        """
        info = {'flavorRef': base.getid(flavor)}
        if disk_config is not None:
            info['OS-DCF:diskConfig'] = disk_config

        return self._action('resize', server, info=info)

    def confirm_resize(self, server):
        """
        Confirm that the resize worked, thus removing the original server.

        :param server: The :class:`Server` (or its ID) to share onto.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('confirmResize', server)

    def revert_resize(self, server):
        """
        Revert a previous resize, switching back to the old server.

        :param server: The :class:`Server` (or its ID) to share onto.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('revertResize', server)

    def create_image(self, server, image_name, metadata=None):
        """
        Snapshot a server.

        :param server: The :class:`Server` (or its ID) to share onto.
        :param image_name: Name to give the snapshot image
        :param metadata: Metadata to give newly-created image entity
        :returns: An instance of novaclient.base.StrWithMeta
                  (The snapshot image's UUID)
        """
        body = {'name': image_name, 'metadata': metadata or {}}
        resp, body = self._action_return_resp_and_body('createImage', server,
                                                       body)
        # The 2.45 microversion returns the image_id in the response body,
        # not as a location header.
        if self.api_version >= api_versions.APIVersion('2.45'):
            image_uuid = body['image_id']
        else:
            location = resp.headers['location']
            image_uuid = location.split('/')[-1]
        return base.StrWithMeta(image_uuid, resp)

    def backup(self, server, backup_name, backup_type, rotation):
        """
        Backup a server instance.

        :param server: The :class:`Server` (or its ID) to share onto.
        :param backup_name: Name of the backup image
        :param backup_type: The backup type, like 'daily' or 'weekly'
        :param rotation: Int parameter representing how many backups to
                        keep around.
        :returns: An instance of novaclient.base.TupleWithMeta if the request
            microversion is < 2.45, otherwise novaclient.base.DictWithMeta.
        """
        body = {'name': backup_name,
                'backup_type': backup_type,
                'rotation': rotation}
        return self._action('createBackup', server, body)

    def set_meta(self, server, metadata):
        """
        Set a server's metadata
        :param server: The :class:`Server` to add metadata to
        :param metadata: A dict of metadata to be added to the server
        """
        body = {'metadata': metadata}
        return self._create("/servers/%s/metadata" % base.getid(server),
                            body, "metadata")

    def set_meta_item(self, server, key, value):
        """
        Updates an item of server metadata
        :param server: The :class:`Server` to add metadata to
        :param key: metadata key to update
        :param value: string value
        """
        body = {'meta': {key: value}}
        return self._update("/servers/%s/metadata/%s" %
                            (base.getid(server), key), body)

    def list_meta(self, server):
        """
        Lists all metadata for a server.
        :param server: The :class:`Server` (or its ID).
        :returns: A dict of metadata.
        """
        resp, body = self.api.client.get("/servers/%s/metadata" %
                                         base.getid(server))
        return base.DictWithMeta(body, resp)

    def get_meta(self, server, key):
        """
        Shows details for a metadata item, by key, for a server.
        :param server: The :class:`Server` (or its ID).
        :param key: metadata key to get
        """
        resp, body = self.api.client.get("/servers/%s/metadata/%s" %
                                         (base.getid(server), key))
        return base.DictWithMeta(body, resp)

    def get_console_output(self, server, length=None):
        """
        Get text console log output from Server.

        :param server: The :class:`Server` (or its ID) whose console output
                        you would like to retrieve.
        :param length: The number of tail loglines you would like to retrieve.
        :returns: An instance of novaclient.base.StrWithMeta or
                  novaclient.base.UnicodeWithMeta
        """
        resp, body = self._action_return_resp_and_body('os-getConsoleOutput',
                                                       server,
                                                       {'length': length})
        return self.convert_into_with_meta(body['output'], resp)

    def delete_meta(self, server, keys):
        """
        Delete metadata from a server

        :param server: The :class:`Server` to add metadata to
        :param keys: A list of metadata keys to delete from the server
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        result = base.TupleWithMeta((), None)
        for k in keys:
            ret = self._delete("/servers/%s/metadata/%s" %
                               (base.getid(server), k))
            result.append_request_ids(ret.request_ids)

        return result

    def _live_migrate(self, server, host, block_migration, disk_over_commit,
                      force):
        """Inner function to abstract changes in live migration API."""
        body = {
            'host': host,
            'block_migration': block_migration,
        }

        if disk_over_commit is not None:
            body['disk_over_commit'] = disk_over_commit

        # NOTE(stephenfin): For some silly reason, we don't set this if it's
        # False, hence why we're not explicitly checking against None
        if force:
            body['force'] = force

        return self._action('os-migrateLive', server, body)

    @api_versions.wraps('2.0', '2.24')
    def live_migrate(self, server, host, block_migration, disk_over_commit):
        """
        Migrates a running instance to a new machine.

        :param server: instance id which comes from nova list.
        :param host: destination host name.
        :param block_migration: if True, do block_migration.
        :param disk_over_commit: if True, allow disk overcommit.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._live_migrate(server, host,
                                  block_migration=block_migration,
                                  disk_over_commit=disk_over_commit,
                                  force=None)

    @api_versions.wraps('2.25', '2.29')
    def live_migrate(self, server, host, block_migration):
        """
        Migrates a running instance to a new machine.

        :param server: instance id which comes from nova list.
        :param host: destination host name.
        :param block_migration: if True, do block_migration, can be set as
                                'auto'
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._live_migrate(server, host,
                                  block_migration=block_migration,
                                  disk_over_commit=None,
                                  force=None)

    @api_versions.wraps('2.30', '2.67')
    def live_migrate(self, server, host, block_migration, force=None):
        """
        Migrates a running instance to a new machine.

        :param server: instance id which comes from nova list.
        :param host: destination host name.
        :param block_migration: if True, do block_migration, can be set as
                                'auto'
        :param force: forces to bypass the scheduler if host is provided.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._live_migrate(server, host,
                                  block_migration=block_migration,
                                  disk_over_commit=None,
                                  force=force)

    @api_versions.wraps('2.68')
    def live_migrate(self, server, host, block_migration):
        """
        Migrates a running instance to a new machine.

        :param server: instance id which comes from nova list.
        :param host: destination host name.
        :param block_migration: if True, do block_migration, can be set as
                                'auto'
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._live_migrate(server, host,
                                  block_migration=block_migration,
                                  disk_over_commit=None,
                                  force=None)

    def reset_state(self, server, state='error'):
        """
        Reset the state of an instance to active or error.

        :param server: ID of the instance to reset the state of.
        :param state: Desired state; either 'active' or 'error'.
                      Defaults to 'error'.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('os-resetState', server, dict(state=state))

    def reset_network(self, server):
        """
        Reset network of an instance.

        :param server: The :class:`Server` for network is to be reset
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('resetNetwork', server)

    def add_security_group(self, server, security_group):
        """
        Add a Security Group to an instance

        :param server: ID of the instance.
        :param security_group: The name of security group to add.
        :returns: An instance of novaclient.base.DictWithMeta
        """
        return self._action('addSecurityGroup', server,
                            {'name': security_group})

    def remove_security_group(self, server, security_group):
        """
        Remove a Security Group to an instance

        :param server: ID of the instance.
        :param security_group: The name of security group to remove.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._action('removeSecurityGroup', server,
                            {'name': security_group})

    def list_security_group(self, server):
        """
        List Security Group(s) of an instance

        :param server: ID of the instance.

        """
        return self._list('/servers/%s/os-security-groups' %
                          base.getid(server), 'security_groups',
                          SecurityGroup)

    def _evacuate(self, server, host, on_shared_storage, password, force):
        """Inner function to abstract changes in evacuate API."""
        body = {}

        if on_shared_storage is not None:
            body['onSharedStorage'] = on_shared_storage

        if host is not None:
            body['host'] = host

        if password is not None:
            body['adminPass'] = password

        if force:
            body['force'] = force

        resp, body = self._action_return_resp_and_body('evacuate', server,
                                                       body)
        return base.TupleWithMeta((resp, body), resp)

    @api_versions.wraps("2.0", "2.13")
    def evacuate(self, server, host=None, on_shared_storage=True,
                 password=None):
        """
        Evacuate a server instance.

        :param server: The :class:`Server` (or its ID) to evacuate to.
        :param host: Name of the target host.
        :param on_shared_storage: Specifies whether instance files located
                        on shared storage
        :param password: string to set as password on the evacuated server.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._evacuate(server, host,
                              on_shared_storage=on_shared_storage,
                              password=password,
                              force=None)

    @api_versions.wraps("2.14", "2.28")
    def evacuate(self, server, host=None, password=None):
        """
        Evacuate a server instance.

        :param server: The :class:`Server` (or its ID) to evacuate to.
        :param host: Name of the target host.
        :param password: string to set as password on the evacuated server.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._evacuate(server, host,
                              on_shared_storage=None,
                              password=password,
                              force=None)

    @api_versions.wraps("2.29", "2.67")
    def evacuate(self, server, host=None, password=None, force=None):
        """
        Evacuate a server instance.

        :param server: The :class:`Server` (or its ID) to evacuate to.
        :param host: Name of the target host.
        :param password: string to set as password on the evacuated server.
        :param force: forces to bypass the scheduler if host is provided.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._evacuate(server, host,
                              on_shared_storage=None,
                              password=password,
                              force=force)

    @api_versions.wraps("2.68")
    def evacuate(self, server, host=None, password=None):
        """
        Evacuate a server instance.

        :param server: The :class:`Server` (or its ID) to evacuate to.
        :param host: Name of the target host.
        :param password: string to set as password on the evacuated server.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._evacuate(server, host,
                              on_shared_storage=None,
                              password=password,
                              force=None)

    def interface_list(self, server):
        """
        List attached network interfaces

        :param server: The :class:`Server` (or its ID) to query.
        """
        return self._list('/servers/%s/os-interface' % base.getid(server),
                          'interfaceAttachments', obj_class=NetworkInterface)

    @api_versions.wraps("2.0", "2.48")
    def interface_attach(self, server, port_id, net_id, fixed_ip):
        """
        Attach a network_interface to an instance.

        :param server: The :class:`Server` (or its ID) to attach to.
        :param port_id: The port to attach.
        """

        body = {'interfaceAttachment': {}}
        if port_id:
            body['interfaceAttachment']['port_id'] = port_id
        if net_id:
            body['interfaceAttachment']['net_id'] = net_id
        if fixed_ip:
            body['interfaceAttachment']['fixed_ips'] = [
                {'ip_address': fixed_ip}]

        return self._create('/servers/%s/os-interface' % base.getid(server),
                            body, 'interfaceAttachment',
                            obj_class=NetworkInterface)

    @api_versions.wraps("2.49")
    def interface_attach(self, server, port_id, net_id, fixed_ip, tag=None):
        """
        Attach a network_interface to an instance.

        :param server: The :class:`Server` (or its ID) to attach to.
        :param port_id: The port to attach.
                        The port_id and net_id parameters are mutually
                        exclusive.
        :param net_id: The ID of the network to attach.
        :param fixed_ip: The fixed IP addresses. If the fixed_ip is specified,
                         the net_id has to be specified at the same time.
        :param tag: The tag.
        """

        body = {'interfaceAttachment': {}}
        if port_id:
            body['interfaceAttachment']['port_id'] = port_id
        if net_id:
            body['interfaceAttachment']['net_id'] = net_id
        if fixed_ip:
            body['interfaceAttachment']['fixed_ips'] = [
                {'ip_address': fixed_ip}]
        if tag:
            body['interfaceAttachment']['tag'] = tag

        return self._create('/servers/%s/os-interface' % base.getid(server),
                            body, 'interfaceAttachment',
                            obj_class=NetworkInterface)

    def interface_detach(self, server, port_id):
        """
        Detach a network_interface from an instance.

        :param server: The :class:`Server` (or its ID) to detach from.
        :param port_id: The port to detach.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._delete('/servers/%s/os-interface/%s' %
                            (base.getid(server), port_id))

    @api_versions.wraps("2.17")
    def trigger_crash_dump(self, server):
        """Trigger crash dump in an instance"""
        return self._action("trigger_crash_dump", server)

    def _action(self, action, server, info=None, **kwargs):
        """
        Perform a server "action" -- reboot/rebuild/resize/etc.
        """
        resp, body = self._action_return_resp_and_body(action, server,
                                                       info=info, **kwargs)
        return self.convert_into_with_meta(body, resp)

    def _action_return_resp_and_body(self, action, server, info=None,
                                     **kwargs):
        """
        Perform a server "action" and return response headers and body
        """
        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        url = '/servers/%s/action' % base.getid(server)
        return self.api.client.post(url, body=body)

    @api_versions.wraps('2.26')
    def tag_list(self, server):
        """
        Get list of tags from an instance.
        """
        resp, body = self.api.client.get(
            "/servers/%s/tags" % base.getid(server))
        return base.ListWithMeta(body['tags'], resp)

    @api_versions.wraps('2.26')
    def delete_tag(self, server, tag):
        """
        Remove single tag from an instance.
        """
        return self._delete("/servers/%s/tags/%s" % (base.getid(server), tag))

    @api_versions.wraps('2.26')
    def delete_all_tags(self, server):
        """
        Remove all tags from an instance.
        """
        return self._delete("/servers/%s/tags" % base.getid(server))

    @api_versions.wraps('2.26')
    def set_tags(self, server, tags):
        """
        Set list of tags to an instance.
        """
        body = {"tags": tags}
        return self._update("/servers/%s/tags" % base.getid(server), body)

    @api_versions.wraps('2.26')
    def add_tag(self, server, tag):
        """
        Add single tag to an instance.
        """
        return self._update(
            "/servers/%s/tags/%s" % (base.getid(server), tag), None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from novaclient.tests.unit.fixture_data import base


class Fixture(base.Fixture):

    base_url = 'limits'
    absolute = {
        "maxTotalRAMSize": 51200,
        "maxServerMeta": 5,
        "maxImageMeta": 5,
        "maxPersonality": 5,
        "maxPersonalitySize": 10240
    }

    def setUp(self):
        super(Fixture, self).setUp()

        get_limits = {
            "limits": {
                "rate": [
                    {
                        "uri": "*",
                        "regex": ".*",
                        "limit": [
                            {
                                "value": 10,
                                "verb": "POST",
                                "remaining": 2,
                                "unit": "MINUTE",
                                "next-available": "2011-12-15T22:42:45Z"
                            },
                            {
                                "value": 10,
                                "verb": "PUT",
                                "remaining": 2,
                                "unit": "MINUTE",
                                "next-available": "2011-12-15T22:42:45Z"
                            },
                            {
                                "value": 100,
                                "verb": "DELETE",
                                "remaining": 100,
                                "unit": "MINUTE",
                                "next-available": "2011-12-15T22:42:45Z"
                            }
                        ]
                    },
                    {
                        "uri": "*/servers",
                        "regex": "^/servers",
                        "limit": [
                            {
                                "verb": "POST",
                                "value": 25,
                                "remaining": 24,
                                "unit": "DAY",
                                "next-available": "2011-12-15T22:42:45Z"
                            }
                        ]
                    }
                ],
                "absolute": self.absolute,
            },
        }

        headers = self.json_headers
        self.requests_mock.get(self.url(),
                               json=get_limits,
                               headers=headers)


class Fixture2_57(Fixture):
    """Fixture data for the 2.57 microversion where personality files are
    deprecated.
    """
    absolute = {
        "maxTotalRAMSize": 51200,
        "maxServerMeta": 5
    }



# Copyright 2011 Denali Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Volume interface
"""
import warnings

from novaclient import api_versions
from novaclient import base


class Volume(base.Resource):
    """
    A volume is an extra block level storage to the OpenStack
    instances.
    """
    NAME_ATTR = 'display_name'

    def __repr__(self):
        return "<Volume: %s>" % self.id


class VolumeManager(base.Manager):
    """
    Manage :class:`Volume` resources. This is really about volume attachments.
    """
    resource_class = Volume

    @staticmethod
    def _get_request_body_for_create(volume_id, device=None, tag=None,
                                     delete_on_termination=False):
        body = {'volumeAttachment': {'volumeId': volume_id}}
        if device is not None:
            body['volumeAttachment']['device'] = device
        if tag is not None:
            body['volumeAttachment']['tag'] = tag
        if delete_on_termination:
            body['volumeAttachment']['delete_on_termination'] = (
                delete_on_termination)

        return body

    @api_versions.wraps("2.0", "2.48")
    def create_server_volume(self, server_id, volume_id, device=None):
        """
        Attach a volume identified by the volume ID to the given server ID

        :param server_id: The ID of the server
        :param volume_id: The ID of the volume to attach.
        :param device: The device name (optional)
        :rtype: :class:`Volume`
        """
        return self._create(
            "/servers/%s/os-volume_attachments" % server_id,
            VolumeManager._get_request_body_for_create(volume_id, device),
            "volumeAttachment")

    @api_versions.wraps("2.49", "2.78")
    def create_server_volume(self, server_id, volume_id, device=None,
                             tag=None):
        """
        Attach a volume identified by the volume ID to the given server ID

        :param server_id: The ID of the server
        :param volume_id: The ID of the volume to attach.
        :param device: The device name (optional)
        :param tag: The tag (optional)
        :rtype: :class:`Volume`
        """
        return self._create(
            "/servers/%s/os-volume_attachments" % server_id,
            VolumeManager._get_request_body_for_create(volume_id, device, tag),
            "volumeAttachment")

    @api_versions.wraps("2.79")
    def create_server_volume(self, server_id, volume_id, device=None,
                             tag=None, delete_on_termination=False):
        """
        Attach a volume identified by the volume ID to the given server ID

        :param server_id: The ID of the server.
        :param volume_id: The ID of the volume to attach.
        :param device: The device name (optional).
        :param tag: The tag (optional).
        :param delete_on_termination: Marked whether to delete the attached
                                      volume when the server is deleted
                                      (optional).
        :rtype: :class:`Volume`
        """
        return self._create(
            "/servers/%s/os-volume_attachments" % server_id,
            VolumeManager._get_request_body_for_create(volume_id, device, tag,
                                                       delete_on_termination),
            "volumeAttachment")

    @api_versions.wraps("2.0", "2.84")
    def update_server_volume(self, server_id, src_volid, dest_volid):
        """
        Swaps the existing volume attachment to point to a new volume.

        Takes a server, a source (attached) volume and a destination volume and
        performs a hypervisor assisted data migration from src to dest volume,
        detaches the original (source) volume and attaches the new destination
        volume. Note that not all backing hypervisor drivers support this
        operation and it may be disabled via policy.


        :param server_id: The ID of the server
        :param source_volume: The ID of the src volume
        :param dest_volume: The ID of the destination volume
        :rtype: :class:`Volume`
        """
        body = {'volumeAttachment': {'volumeId': dest_volid}}
        return self._update("/servers/%s/os-volume_attachments/%s" %
                            (server_id, src_volid,),
                            body, "volumeAttachment")

    @api_versions.wraps("2.85")
    def update_server_volume(self, server_id, src_volid, dest_volid,
                             delete_on_termination=None):
        """
        Swaps the existing volume attachment to point to a new volume.

        Takes a server, a source (attached) volume and a destination volume and
        performs a hypervisor assisted data migration from src to dest volume,
        detaches the original (source) volume and attaches the new destination
        volume. Note that not all backing hypervisor drivers support this
        operation and it may be disabled via policy.


        :param server_id: The ID of the server
        :param source_volume: The ID of the src volume
        :param dest_volume: The ID of the destination volume
        :param delete_on_termination: Marked whether to delete the attached
                                      volume when the server is deleted
                                      (optional).
        :rtype: :class:`Volume`
        """
        body = {'volumeAttachment': {'volumeId': dest_volid}}
        if delete_on_termination is not None:
            body['volumeAttachment']['delete_on_termination'] = (
                delete_on_termination)
        return self._update("/servers/%s/os-volume_attachments/%s" %
                            (server_id, src_volid),
                            body, "volumeAttachment")

    def get_server_volume(self, server_id, volume_id=None, attachment_id=None):
        """
        Get the volume identified by the volume ID, that is attached to
        the given server ID

        :param server_id: The ID of the server
        :param volume_id: The ID of the volume to attach
        :rtype: :class:`Volume`
        """

        if attachment_id is not None and volume_id is not None:
            raise TypeError("You cannot specify both volume_id "
                            "and attachment_id arguments.")

        elif attachment_id is not None:
            warnings.warn("attachment_id argument "
                          "of volumes.get_server_volume "
                          "method is deprecated in favor "
                          "of volume_id.")
            volume_id = attachment_id

        if volume_id is None:
            raise TypeError("volume_id is required argument.")

        return self._get("/servers/%s/os-volume_attachments/%s" % (server_id,
                         volume_id,), "volumeAttachment")

    def get_server_volumes(self, server_id):
        """
        Get a list of all the attached volumes for the given server ID

        :param server_id: The ID of the server
        :rtype: list of :class:`Volume`
        """
        return self._list("/servers/%s/os-volume_attachments" % server_id,
                          "volumeAttachments")

    def delete_server_volume(self, server_id, volume_id=None,
                             attachment_id=None):
        """
        Detach a volume identified by the volume ID from the given server

        :param server_id: The ID of the server
        :param volume_id: The ID of the volume to attach
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if attachment_id is not None and volume_id is not None:
            raise TypeError("You cannot specify both volume_id "
                            "and attachment_id arguments.")

        elif attachment_id is not None:
            warnings.warn("attachment_id argument "
                          "of volumes.delete_server_volume "
                          "method is deprecated in favor "
                          "of volume_id.")
            volume_id = attachment_id

        if volume_id is None:
            raise TypeError("volume_id is required argument.")

        return self._delete("/servers/%s/os-volume_attachments/%s" %
                            (server_id, volume_id,))



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Aggregate interface."""

from novaclient import api_versions
from novaclient import base


class Aggregate(base.Resource):
    """An aggregates is a collection of compute hosts."""

    def __repr__(self):
        return "<Aggregate: %s>" % self.id

    def update(self, values):
        """Update the name and/or availability zone."""
        return self.manager.update(self, values)

    def add_host(self, host):
        return self.manager.add_host(self, host)

    def remove_host(self, host):
        return self.manager.remove_host(self, host)

    def set_metadata(self, metadata):
        return self.manager.set_metadata(self, metadata)

    def delete(self):
        """
        Delete the own aggregate.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.delete(self)

    @api_versions.wraps("2.81")
    def cache_images(self, images):
        return self.manager.cache_images(self, images)


class AggregateManager(base.ManagerWithFind):
    resource_class = Aggregate

    def list(self):
        """Get a list of os-aggregates."""
        return self._list('/os-aggregates', 'aggregates')

    def create(self, name, availability_zone):
        """Create a new aggregate."""
        body = {'aggregate': {'name': name,
                              'availability_zone': availability_zone}}
        return self._create('/os-aggregates', body, 'aggregate')

    def get(self, aggregate):
        """Get details of the specified aggregate."""
        return self._get('/os-aggregates/%s' % (base.getid(aggregate)),
                         "aggregate")

    # NOTE:(dtroyer): utils.find_resource() uses manager.get() but we need to
    #                 keep the API backward compatible
    def get_details(self, aggregate):
        """Get details of the specified aggregate."""
        return self.get(aggregate)

    def update(self, aggregate, values):
        """Update the name and/or availability zone."""
        body = {'aggregate': values}
        return self._update("/os-aggregates/%s" % base.getid(aggregate),
                            body,
                            "aggregate")

    def add_host(self, aggregate, host):
        """Add a host into the Host Aggregate."""
        body = {'add_host': {'host': host}}
        return self._create("/os-aggregates/%s/action" % base.getid(aggregate),
                            body, "aggregate")

    def remove_host(self, aggregate, host):
        """Remove a host from the Host Aggregate."""
        body = {'remove_host': {'host': host}}
        return self._create("/os-aggregates/%s/action" % base.getid(aggregate),
                            body, "aggregate")

    def set_metadata(self, aggregate, metadata):
        """Set aggregate metadata, replacing the existing metadata."""
        body = {'set_metadata': {'metadata': metadata}}
        return self._create("/os-aggregates/%s/action" % base.getid(aggregate),
                            body, "aggregate")

    def delete(self, aggregate):
        """
        Delete the specified aggregate.

        :param aggregate: The aggregate to delete
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._delete('/os-aggregates/%s' % (base.getid(aggregate)))

    @api_versions.wraps("2.81")
    def cache_images(self, aggregate, images):
        """
        Request images be cached on a given aggregate.

        :param aggregate: The aggregate to target
        :param images: A list of image IDs to request caching
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        body = {
            'cache': [{'id': base.getid(image)} for image in images],
        }
        resp, body = self.api.client.post(
            "/os-aggregates/%s/images" % base.getid(aggregate),
            body=body)
        return self.convert_into_with_meta(body, resp)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient.tests.unit.fixture_data import client
from novaclient.tests.unit.fixture_data import limits as data
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes
from novaclient.v2 import limits


class LimitsTest(utils.FixturedTestCase):

    client_fixture_class = client.V1
    data_fixture_class = data.Fixture
    supports_image_meta = True   # 2.39 deprecates maxImageMeta
    supports_personality = True  # 2.57 deprecates maxPersonality*

    def test_get_limits(self):
        obj = self.cs.limits.get()
        self.assert_request_id(obj, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/limits')
        self.assertIsInstance(obj, limits.Limits)

    def test_get_limits_for_a_tenant(self):
        obj = self.cs.limits.get(tenant_id=1234)
        self.assert_request_id(obj, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/limits?tenant_id=1234')
        self.assertIsInstance(obj, limits.Limits)

    def test_absolute_limits_reserved(self):
        obj = self.cs.limits.get(reserved=True)
        self.assert_request_id(obj, fakes.FAKE_REQUEST_ID_LIST)

        expected = [
            limits.AbsoluteLimit("maxTotalRAMSize", 51200),
            limits.AbsoluteLimit("maxServerMeta", 5)
        ]
        if self.supports_image_meta:
            expected.append(limits.AbsoluteLimit("maxImageMeta", 5))
        if self.supports_personality:
            expected.extend([
                limits.AbsoluteLimit("maxPersonality", 5),
                limits.AbsoluteLimit("maxPersonalitySize", 10240)])

        self.assert_called('GET', '/limits?reserved=1')
        abs_limits = list(obj.absolute)
        self.assertEqual(len(abs_limits), len(expected))

        for limit in abs_limits:
            self.assertIn(limit, expected)

    def test_rate_absolute_limits(self):
        obj = self.cs.limits.get()
        self.assert_request_id(obj, fakes.FAKE_REQUEST_ID_LIST)

        expected = (
            limits.RateLimit('POST', '*', '.*', 10, 2, 'MINUTE',
                             '2011-12-15T22:42:45Z'),
            limits.RateLimit('PUT', '*', '.*', 10, 2, 'MINUTE',
                             '2011-12-15T22:42:45Z'),
            limits.RateLimit('DELETE', '*', '.*', 100, 100, 'MINUTE',
                             '2011-12-15T22:42:45Z'),
            limits.RateLimit('POST', '*/servers', '^/servers', 25, 24, 'DAY',
                             '2011-12-15T22:42:45Z'),
        )

        rate_limits = list(obj.rate)
        self.assertEqual(len(rate_limits), len(expected))

        for limit in rate_limits:
            self.assertIn(limit, expected)

        expected = [
            limits.AbsoluteLimit("maxTotalRAMSize", 51200),
            limits.AbsoluteLimit("maxServerMeta", 5)
        ]
        if self.supports_image_meta:
            expected.append(limits.AbsoluteLimit("maxImageMeta", 5))
        if self.supports_personality:
            expected.extend([
                limits.AbsoluteLimit("maxPersonality", 5),
                limits.AbsoluteLimit("maxPersonalitySize", 10240)])

        abs_limits = list(obj.absolute)
        self.assertEqual(len(abs_limits), len(expected))

        for limit in abs_limits:
            self.assertIn(limit, expected)


class LimitsTest2_57(LimitsTest):
    data_fixture_class = data.Fixture2_57
    supports_image_meta = False
    supports_personality = False

    def setUp(self):
        super(LimitsTest2_57, self).setUp()
        self.cs.api_version = api_versions.APIVersion('2.57')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from urllib import parse

import fixtures

from novaclient.tests.unit.v2 import fakes

COMPUTE_URL = 'http://compute.host'


class Fixture(fixtures.Fixture):

    base_url = None
    json_headers = {'Content-Type': 'application/json',
                    'x-openstack-request-id': fakes.FAKE_REQUEST_ID}

    def __init__(self, requests_mock, compute_url=COMPUTE_URL):
        super(Fixture, self).__init__()
        self.requests_mock = requests_mock
        self.compute_url = compute_url

    def url(self, *args, **kwargs):
        url_args = [self.compute_url]

        if self.base_url:
            url_args.append(self.base_url)

        url = '/'.join(str(a).strip('/') for a in tuple(url_args) + args)

        if kwargs:
            url += '?%s' % parse.urlencode(kwargs, doseq=True)

        return url



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient import base
from novaclient.tests.unit.fixture_data import client
from novaclient.tests.unit.fixture_data import server_migrations as data
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes
from novaclient.v2 import server_migrations


class ServerMigrationsTest(utils.FixturedTestCase):
    client_fixture_class = client.V1
    data_fixture_class = data.Fixture

    def setUp(self):
        super(ServerMigrationsTest, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.22")

    def test_live_migration_force_complete(self):
        body = {'force_complete': None}
        self.cs.server_migrations.live_migrate_force_complete(1234, 1)
        self.assert_called('POST', '/servers/1234/migrations/1/action', body)


class ServerMigrationsTestV223(ServerMigrationsTest):

    migration = {
        "created_at": "2016-01-29T13:42:02.000000",
        "dest_compute": "compute2",
        "dest_host": "1.2.3.4",
        "dest_node": "node2",
        "id": 1,
        "server_uuid": "4cfba335-03d8-49b2-8c52-e69043d1e8fe",
        "source_compute": "compute1",
        "source_node": "node1",
        "status": "running",
        "memory_total_bytes": 123456,
        "memory_processed_bytes": 12345,
        "memory_remaining_bytes": 120000,
        "disk_total_bytes": 234567,
        "disk_processed_bytes": 23456,
        "disk_remaining_bytes": 230000,
        "updated_at": "2016-01-29T13:42:02.000000"
    }

    def setUp(self):
        super(ServerMigrationsTestV223, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.23")

    def test_list_migrations(self):
        ml = self.cs.server_migrations.list(1234)

        self.assertIsInstance(ml, base.ListWithMeta)
        self.assert_request_id(ml, fakes.FAKE_REQUEST_ID_LIST)
        for k in self.migration:
            self.assertEqual(self.migration[k], getattr(ml[0], k))

        self.assert_called('GET', '/servers/1234/migrations')

    def test_get_migration(self):
        migration = self.cs.server_migrations.get(1234, 1)

        self.assertIsInstance(migration, server_migrations.ServerMigration)
        for k in migration._info:
            self.assertEqual(self.migration[k], migration._info[k])
        self.assert_request_id(migration, fakes.FAKE_REQUEST_ID_LIST)

        self.assert_called('GET', '/servers/1234/migrations/1')


class ServerMigrationsTestV224(ServerMigrationsTest):
    def setUp(self):
        super(ServerMigrationsTestV224, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.24")

    def test_live_migration_abort(self):
        self.cs.server_migrations.live_migration_abort(1234, 1)
        self.assert_called('DELETE', '/servers/1234/migrations/1')



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient import base


class QuotaClassSet(base.Resource):

    def update(self, *args, **kwargs):
        return self.manager.update(self.id, *args, **kwargs)


class QuotaClassSetManager(base.Manager):
    resource_class = QuotaClassSet

    def get(self, class_name):
        return self._get("/os-quota-class-sets/%s" % (class_name),
                         "quota_class_set")

    def _update_body(self, **kwargs):
        return {'quota_class_set': kwargs}

    # NOTE(mriedem): Before 2.50 the resources you could update was just a
    # kwargs dict and not validated on the client-side, only on the API server
    # side.
    @api_versions.wraps("2.0", "2.49")
    def update(self, class_name, **kwargs):
        body = self._update_body(**kwargs)

        for key in list(body['quota_class_set']):
            if body['quota_class_set'][key] is None:
                body['quota_class_set'].pop(key)

        return self._update('/os-quota-class-sets/%s' % (class_name),
                            body,
                            'quota_class_set')

    # NOTE(mriedem): 2.50 does strict validation of the resources you can
    # specify since the network-related resources are blocked in 2.50.
    @api_versions.wraps("2.50", "2.56")
    def update(self, class_name, instances=None, cores=None, ram=None,
               metadata_items=None, injected_files=None,
               injected_file_content_bytes=None, injected_file_path_bytes=None,
               key_pairs=None, server_groups=None, server_group_members=None):
        resources = {}
        if instances is not None:
            resources['instances'] = instances
        if cores is not None:
            resources['cores'] = cores
        if ram is not None:
            resources['ram'] = ram
        if metadata_items is not None:
            resources['metadata_items'] = metadata_items
        if injected_files is not None:
            resources['injected_files'] = injected_files
        if injected_file_content_bytes is not None:
            resources['injected_file_content_bytes'] = (
                injected_file_content_bytes)
        if injected_file_path_bytes is not None:
            resources['injected_file_path_bytes'] = injected_file_path_bytes
        if key_pairs is not None:
            resources['key_pairs'] = key_pairs
        if server_groups is not None:
            resources['server_groups'] = server_groups
        if server_group_members is not None:
            resources['server_group_members'] = server_group_members

        body = {'quota_class_set': resources}
        return self._update('/os-quota-class-sets/%s' % class_name, body,
                            'quota_class_set')

    # NOTE(mriedem): 2.57 deprecates the usage of injected_files,
    # injected_file_content_bytes and injected_file_path_bytes so those
    # kwargs are removed.
    @api_versions.wraps("2.57")
    def update(self, class_name, instances=None, cores=None, ram=None,
               metadata_items=None, key_pairs=None, server_groups=None,
               server_group_members=None):
        resources = {}
        if instances is not None:
            resources['instances'] = instances
        if cores is not None:
            resources['cores'] = cores
        if ram is not None:
            resources['ram'] = ram
        if metadata_items is not None:
            resources['metadata_items'] = metadata_items
        if key_pairs is not None:
            resources['key_pairs'] = key_pairs
        if server_groups is not None:
            resources['server_groups'] = server_groups
        if server_group_members is not None:
            resources['server_group_members'] = server_group_members

        body = {'quota_class_set': resources}
        return self._update('/os-quota-class-sets/%s' % class_name, body,
                            'quota_class_set')



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient import utils


@api_versions.wraps("2.10", "2.20")
def do_fake_action():
    return 1


@api_versions.wraps("2.21", "2.30")
def do_fake_action():
    return 2


@api_versions.wraps("2.0")
def do_another_fake_action():
    return 0


@utils.arg(
    '--foo',
    start_version='2.1',
    end_version='2.2')
@utils.arg(
    '--bar',
    start_version='2.3',
    end_version='2.4')
def do_fake_action2():
    return 3


@utils.arg(
    '--foo',
    help='first foo',
    start_version='2.10',
    end_version='2.20')
@utils.arg(
    '--foo',
    help='second foo',
    start_version='2.21')
def do_fake_action3():
    return 3



# Copyright (C) 2014, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
External event triggering for servers, not to be used by users.
"""

from novaclient import base


class Event(base.Resource):
    def __repr__(self):
        return "<Event: %s>" % self.name


class ServerExternalEventManager(base.Manager):
    resource_class = Event

    def create(self, events):
        """Create one or more server events.

        :param:events: A list of dictionaries containing 'server_uuid', 'name',
                       'status', and 'tag' (which may be absent)
        """

        body = {'events': events}
        return self._create('/os-server-external-events', body, 'events',
                            return_raw=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from novaclient import api_versions
from novaclient.tests.unit import fakes
from novaclient.tests.unit.fixture_data import base
from novaclient.tests.unit.v2 import fakes as v2_fakes


class Base(base.Fixture):

    base_url = 'servers'

    def setUp(self):
        super(Base, self).setUp()

        get_servers = {
            "servers": [
                {'id': 1234, 'name': 'sample-server'},
                {'id': 5678, 'name': 'sample-server2'}
            ]
        }

        self.requests_mock.get(self.url(),
                               json=get_servers,
                               headers=self.json_headers)

        self.server_1234 = {
            "id": 1234,
            "name": "sample-server",
            "image": {
                "id": 2,
                "name": "sample image",
            },
            "flavor": {
                "id": 1,
                "name": "256 MiB Server",
            },
            "hostId": "e4d909c290d0fb1ca068ffaddf22cbd0",
            "status": "BUILD",
            "progress": 60,
            "addresses": {
                "public": [
                    {
                        "version": 4,
                        "addr": "1.2.3.4",
                    },
                    {
                        "version": 4,
                        "addr": "5.6.7.8",
                    }],
                "private": [{
                    "version": 4,
                    "addr": "10.11.12.13",
                }],
            },
            "metadata": {
                "Server Label": "Web Head 1",
                "Image Version": "2.1"
            },
            "OS-EXT-SRV-ATTR:host": "computenode1",
            "security_groups": [{
                'id': 1, 'name': 'securitygroup1',
                'description': 'FAKE_SECURITY_GROUP',
                'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7'
            }],
            "OS-EXT-MOD:some_thing": "mod_some_thing_value",
        }

        self.server_5678 = {
            "id": 5678,
            "name": "sample-server2",
            "image": {
                "id": 2,
                "name": "sample image",
            },
            "flavor": {
                "id": 1,
                "name": "256 MiB Server",
            },
            "hostId": "9e107d9d372bb6826bd81d3542a419d6",
            "status": "ACTIVE",
            "addresses": {
                "public": [
                    {
                        "version": 4,
                        "addr": "4.5.6.7",
                    },
                    {
                        "version": 4,
                        "addr": "5.6.9.8",
                    }],
                "private": [{
                    "version": 4,
                    "addr": "10.13.12.13",
                }],
            },
            "metadata": {
                "Server Label": "DB 1"
            },
            "OS-EXT-SRV-ATTR:host": "computenode2",
            "security_groups": [
                {
                    'id': 1, 'name': 'securitygroup1',
                    'description': 'FAKE_SECURITY_GROUP',
                    'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7'
                },
                {
                    'id': 2, 'name': 'securitygroup2',
                    'description': 'ANOTHER_FAKE_SECURITY_GROUP',
                    'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7'
                }],
        }

        self.server_9012 = {
            "id": 9012,
            "name": "sample-server3",
            "image": "",
            "flavor": {
                "id": 1,
                "name": "256 MiB Server",
            },
            "hostId": "9e107d9d372bb6826bd81d3542a419d6",
            "status": "ACTIVE",
            "addresses": {
                "public": [
                    {
                        "version": 4,
                        "addr": "4.5.6.7",
                    },
                    {
                        "version": 4,
                        "addr": "5.6.9.8",
                    }],
                "private": [{
                    "version": 4,
                    "addr": "10.13.12.13",
                }],
            },
            "metadata": {
                "Server Label": "DB 1"
            }
        }

        servers = [self.server_1234, self.server_5678, self.server_9012]
        get_servers_detail = {"servers": servers}

        self.requests_mock.get(self.url('detail'),
                               json=get_servers_detail,
                               headers=self.json_headers)

        self.requests_mock.get(
            self.url('detail', marker=self.server_1234["id"]),
            json={"servers": [self.server_1234, self.server_5678]},
            headers=self.json_headers, complete_qs=True)

        self.requests_mock.get(
            self.url('detail', marker=self.server_5678["id"]),
            json={"servers": []},
            headers=self.json_headers, complete_qs=True)

        self.server_1235 = self.server_1234.copy()
        self.server_1235['id'] = 1235
        self.server_1235['status'] = 'error'
        self.server_1235['fault'] = {'message': 'something went wrong!'}

        for s in servers + [self.server_1235]:
            self.requests_mock.get(self.url(s['id']),
                                   json={'server': s},
                                   headers=self.json_headers)

        for s in (1234, 5678):
            self.requests_mock.delete(self.url(s),
                                      status_code=202,
                                      headers=self.json_headers)

        for k in ('test_key', 'key1', 'key2'):
            self.requests_mock.delete(self.url(1234, 'metadata', k),
                                      status_code=204,
                                      headers=self.json_headers)

        metadata1 = {'metadata': {'test_key': 'test_value'}}
        self.requests_mock.post(self.url(1234, 'metadata'),
                                json=metadata1,
                                headers=self.json_headers)
        self.requests_mock.put(self.url(1234, 'metadata', 'test_key'),
                               json=metadata1,
                               headers=self.json_headers)

        self.diagnostic = {'data': 'Fake diagnostics'}

        metadata2 = {'metadata': {'key1': 'val1'}}
        for u in ('uuid1', 'uuid2', 'uuid3', 'uuid4'):
            self.requests_mock.post(self.url(u, 'metadata'),
                                    json=metadata2, status_code=204)
            self.requests_mock.delete(self.url(u, 'metadata', 'key1'),
                                      json=self.diagnostic,
                                      headers=self.json_headers)
        metadata3 = {'meta': {
            'Server Label': 'Web Head 1'
        }}
        self.requests_mock.get(self.url(1234, 'metadata', 'Server Label'),
                               json=metadata3,
                               headers=self.json_headers)
        metadata4 = {'metadata': {
            'Server Label': 'Web Head 1',
            'Image Version': '2.1'
        }}
        self.requests_mock.get(self.url(1234, 'metadata'),
                               json=metadata4,
                               headers=self.json_headers)

        get_security_groups = {
            "security_groups": [{
                'id': 1,
                'name': 'securitygroup1',
                'description': 'FAKE_SECURITY_GROUP',
                'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7',
                'rules': []}]
        }

        self.requests_mock.get(self.url('1234', 'os-security-groups'),
                               json=get_security_groups,
                               headers=self.json_headers)

        self.requests_mock.post(self.url(),
                                json=self.post_servers,
                                headers=self.json_headers)

        self.requests_mock.post(self.url('1234', 'remote-consoles'),
                                json=self.post_servers_1234_remote_consoles,
                                headers=self.json_headers)

        self.requests_mock.post(self.url('1234', 'action'),
                                json=self.post_servers_1234_action,
                                headers=self.json_headers)

        get_os_interface = {
            "interfaceAttachments": [
                {
                    "port_state": "ACTIVE",
                    "net_id": "net-id-1",
                    "port_id": "port-id-1",
                    "mac_address": "aa:bb:cc:dd:ee:ff",
                    "fixed_ips": [{"ip_address": "1.2.3.4"}],
                },
                {
                    "port_state": "ACTIVE",
                    "net_id": "net-id-1",
                    "port_id": "port-id-1",
                    "mac_address": "aa:bb:cc:dd:ee:ff",
                    "fixed_ips": [{"ip_address": "1.2.3.4"}],
                }
            ]
        }

        self.requests_mock.get(self.url('1234', 'os-interface'),
                               json=get_os_interface,
                               headers=self.json_headers)

        interface_data = {'interfaceAttachment': {}}
        self.requests_mock.post(self.url('1234', 'os-interface'),
                                json=interface_data,
                                headers=self.json_headers)

        def put_servers_1234(request, context):
            body = request.json()
            assert list(body) == ['server']
            fakes.assert_has_keys(body['server'],
                                  optional=['name', 'adminPass'])
            return request.body

        self.requests_mock.put(self.url(1234),
                               text=put_servers_1234,
                               status_code=204,
                               headers=self.json_headers)

        #
        # Server password
        #

        self.requests_mock.delete(self.url(1234, 'os-server-password'),
                                  status_code=202,
                                  headers=self.json_headers)
        #
        # Server tags
        #

        self.requests_mock.get(self.url(1234, 'tags'),
                               json={'tags': ['tag1', 'tag2']},
                               headers=self.json_headers)

        self.requests_mock.get(self.url(1234, 'tags', 'tag'),
                               status_code=204,
                               headers=self.json_headers)

        self.requests_mock.delete(self.url(1234, 'tags', 'tag'),
                                  status_code=204,
                                  headers=self.json_headers)

        self.requests_mock.delete(self.url(1234, 'tags'),
                                  status_code=204,
                                  headers=self.json_headers)

        def put_server_tag(request, context):
            assert request.text is None
            context.status_code = 201
            return None

        self.requests_mock.put(self.url(1234, 'tags', 'tag'),
                               json=put_server_tag,
                               headers=self.json_headers)

        def put_server_tags(request, context):
            body = request.json()
            assert list(body) == ['tags']
            return body

        self.requests_mock.put(self.url(1234, 'tags'),
                               json=put_server_tags,
                               headers=self.json_headers)


class V1(Base):

    def setUp(self):
        super(V1, self).setUp()

        #
        # Server Addresses
        #

        add = self.server_1234['addresses']
        self.requests_mock.get(self.url(1234, 'ips'),
                               json={'addresses': add},
                               headers=self.json_headers)

        self.requests_mock.get(self.url(1234, 'ips', 'public'),
                               json={'public': add['public']},
                               headers=self.json_headers)

        self.requests_mock.get(self.url(1234, 'ips', 'private'),
                               json={'private': add['private']},
                               headers=self.json_headers)

        self.requests_mock.delete(self.url(1234, 'ips', 'public', '1.2.3.4'),
                                  status_code=202)

        self.requests_mock.get(self.url('1234', 'diagnostics'),
                               json=self.diagnostic,
                               headers=self.json_headers)

        self.requests_mock.delete(self.url('1234', 'os-interface', 'port-id'),
                                  headers=self.json_headers)

        self.requests_mock.get(self.url('1234', 'topology'),
                               json=v2_fakes.SERVER_TOPOLOGY,
                               headers=self.json_headers)

        # Testing with the following password and key
        #
        # Clear password: FooBar123
        #
        # RSA Private Key: novaclient/tests/unit/idfake.pem
        #
        # Encrypted password
        # OIuEuQttO8Rk93BcKlwHQsziDAnkAm/V6V8VPToA8ZeUaUBWwS0gwo2K6Y61Z96r
        # qG447iRz0uTEEYq3RAYJk1mh3mMIRVl27t8MtIecR5ggVVbz1S9AwXJQypDKl0ho
        # QFvhCBcMWPohyGewDJOhDbtuN1IoFI9G55ZvFwCm5y7m7B2aVcoLeIsJZE4PLsIw
        # /y5a6Z3/AoJZYGG7IH5WN88UROU3B9JZGFB2qtPLQTOvDMZLUhoPRIJeHiVSlo1N
        # tI2/++UsXVg3ow6ItqCJGgdNuGG5JB+bslDHWPxROpesEIHdczk46HCpHQN8f1sk
        # Hi/fmZZNQQqj1Ijq0caOIw==

        get_server_password = {
            'password':
            'OIuEuQttO8Rk93BcKlwHQsziDAnkAm/V6V8VPToA8ZeUaUBWwS0gwo2K6Y61Z96r'
            'qG447iRz0uTEEYq3RAYJk1mh3mMIRVl27t8MtIecR5ggVVbz1S9AwXJQypDKl0ho'
            'QFvhCBcMWPohyGewDJOhDbtuN1IoFI9G55ZvFwCm5y7m7B2aVcoLeIsJZE4PLsIw'
            '/y5a6Z3/AoJZYGG7IH5WN88UROU3B9JZGFB2qtPLQTOvDMZLUhoPRIJeHiVSlo1N'
            'tI2/++UsXVg3ow6ItqCJGgdNuGG5JB+bslDHWPxROpesEIHdczk46HCpHQN8f1sk'
            'Hi/fmZZNQQqj1Ijq0caOIw=='}
        self.requests_mock.get(self.url(1234, 'os-server-password'),
                               json=get_server_password,
                               headers=self.json_headers)

    def post_servers(self, request, context):
        body = request.json()
        context.status_code = 202
        assert (set(body.keys()) <=
                set(['server', 'os:scheduler_hints']))
        fakes.assert_has_keys(body['server'],
                              required=['name', 'imageRef', 'flavorRef'],
                              optional=['metadata', 'personality'])
        if 'personality' in body['server']:
            for pfile in body['server']['personality']:
                fakes.assert_has_keys(pfile, required=['path', 'contents'])
        if ('return_reservation_id' in body['server'].keys() and
                body['server']['return_reservation_id']):
            return {'reservation_id': 'r-3fhpjulh'}
        if body['server']['name'] == 'some-bad-server':
            body = self.server_1235
        else:
            body = self.server_1234

        return {'server': body}

    def post_servers_1234_remote_consoles(self, request, context):
        _body = ''
        body = request.json()
        context.status_code = 202
        assert len(body.keys()) == 1
        assert 'remote_console' in body.keys()
        assert 'protocol' in body['remote_console'].keys()
        protocol = body['remote_console']['protocol']

        _body = {'protocol': protocol, 'type': 'novnc',
                 'url': 'http://example.com:6080/vnc_auto.html?token=XYZ'}

        return {'remote_console': _body}

    def post_servers_1234_action(self, request, context):
        _body = ''
        body = request.json()
        context.status_code = 202
        assert len(body.keys()) == 1
        action = list(body)[0]
        api_version = api_versions.APIVersion(
            request.headers.get('X-OpenStack-Nova-API-Version', '2.1'))

        if v2_fakes.FakeSessionClient.check_server_actions(body):
            # NOTE(snikitin): No need to do any operations here. This 'pass'
            # is needed to avoid AssertionError in the last 'else' statement
            # if we found 'action' in method check_server_actions and
            # raise AssertionError if we didn't find 'action' at all.
            pass
        elif action == 'os-migrateLive':
            # Fixme(eliqiao): body of os-migrateLive changes from v2.25
            # but we can not specify version in data_fixture now and this is
            # V1 data, so just let it pass
            pass
        elif action == 'migrate':
            return None
        elif action == 'lock':
            return None
        elif action == 'unshelve':
            if api_version >= api_versions.APIVersion("2.91"):
                # In 2.91 and above, we allow body to be one of these:
                # {'unshelve': None}
                # {'unshelve': {'availability_zone': <string>}}
                # {'unshelve': {'availability_zone': None}}   (Unpin az)
                # {'unshelve': {'host': <fqdn>}}
                # {'unshelve': {'availability_zone': <string>, 'host': <fqdn>}}
                # {'unshelve': {'availability_zone': None, 'host': <fqdn>}}
                if body[action] is not None:
                    for key in body[action].keys():
                        key in ['availability_zone', 'host']
            return None
        elif action == 'rebuild':
            body = body[action]
            adminPass = body.get('adminPass', 'randompassword')
            assert 'imageRef' in body
            _body = self.server_1234.copy()
            _body['adminPass'] = adminPass
        elif action == 'confirmResize':
            assert body[action] is None
            # This one method returns a different response code
            context.status_code = 204
            return None
        elif action == 'rescue':
            if body[action]:
                keys = set(body[action].keys())
                assert not (keys - set(['adminPass', 'rescue_image_ref']))
            else:
                assert body[action] is None
            _body = {'adminPass': 'RescuePassword'}
        elif action == 'createImage':
            assert set(body[action].keys()) == set(['name', 'metadata'])
            if api_version >= api_versions.APIVersion('2.45'):
                return {'image_id': '456'}
            context.headers['location'] = "http://blah/images/456"
        elif action == 'createBackup':
            assert set(body[action].keys()) == set(['name', 'backup_type',
                                                    'rotation'])
            if api_version >= api_versions.APIVersion('2.45'):
                return {'image_id': '456'}
            context.headers['location'] = "http://blah/images/456"
        elif action == 'os-getConsoleOutput':
            assert list(body[action]) == ['length']
            context.status_code = 202
            return {'output': 'foo'}
        elif action == 'os-getSerialConsole':
            assert list(body[action]) == ['type']
        elif action == 'evacuate':
            keys = list(body[action])
            if 'adminPass' in keys:
                keys.remove('adminPass')
            assert 'host' in keys
        else:
            raise AssertionError("Unexpected server action: %s" % action)
        return {'server': _body}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient.tests.functional.v2.legacy import test_flavor_access


class TestFlvAccessNovaClientV27(test_flavor_access.TestFlvAccessNovaClient):
    """Check that an attempt to grant an access to a public flavor
      for the given tenant fails with Conflict error in accordance with
      2.7 microversion REST API History
    """

    COMPUTE_API_VERSION = "2.7"

    def test_add_access_public_flavor(self):
        flv_name = self.name_generate()
        self.nova('flavor-create %s auto 512 1 1' % flv_name)
        self.addCleanup(self.nova, 'flavor-delete %s' % flv_name)
        output = self.nova('flavor-access-add %s %s' %
                           (flv_name, self.project_id),
                           fail_ok=True, merge_stderr=True)
        self.assertIn("ERROR (Conflict): "
                      "Can not add access to a public flavor. (HTTP 409) ",
                      output)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import os
import re
import textwrap
import time
from urllib import parse

from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from oslo_utils import uuidutils
import prettytable

from novaclient import exceptions
from novaclient.i18n import _


VALID_KEY_REGEX = re.compile(r"[\w\.\- :]+$", re.UNICODE)


def env(*args, **kwargs):
    """Returns the first environment variable set.

    If all are empty, defaults to '' or keyword arg `default`.
    """
    for arg in args:
        value = os.environ.get(arg)
        if value:
            return value
    return kwargs.get('default', '')


def get_service_type(f):
    """Retrieves service type from function."""
    return getattr(f, 'service_type', None)


def unauthenticated(func):
    """Adds 'unauthenticated' attribute to decorated function.

    Usage:

    >>> @unauthenticated
    ... def mymethod(f):
    ...     pass
    """
    func.unauthenticated = True
    return func


def isunauthenticated(func):
    """Checks if the function does not require authentication.

    Mark such functions with the `@unauthenticated` decorator.

    :returns: bool
    """
    return getattr(func, 'unauthenticated', False)


def arg(*args, **kwargs):
    """Decorator for CLI args.

    Example:

    >>> @arg("name", help="Name of the new entity")
    ... def entity_create(args):
    ...     pass
    """
    def _decorator(func):
        add_arg(func, *args, **kwargs)
        return func
    return _decorator


def add_arg(func, *args, **kwargs):
    """Bind CLI arguments to a shell.py `do_foo` function."""

    if not hasattr(func, 'arguments'):
        func.arguments = []

    # NOTE(sirp): avoid dups that can occur when the module is shared across
    # tests.
    if (args, kwargs) not in func.arguments:
        # Because of the semantics of decorator composition if we just append
        # to the options list positional options will appear to be backwards.
        func.arguments.insert(0, (args, kwargs))


def service_type(stype):
    """Adds 'service_type' attribute to decorated function.

    Usage:

    .. code-block:: python

       @service_type('volume')
       def mymethod(f):
       ...
    """
    def inner(f):
        f.service_type = stype
        return f
    return inner


def pretty_choice_list(values):
    return ', '.join("'%s'" % x for x in values)


def pretty_choice_dict(values):
    """Returns a formatted dict as 'key=value'."""
    return pretty_choice_list(
        ['%s=%s' % (k, values[k]) for k in sorted(values)])


def print_list(objs, fields, formatters={}, sortby_index=None):
    if sortby_index is None:
        sortby = None
    else:
        sortby = fields[sortby_index]
    mixed_case_fields = ['serverId']
    pt = prettytable.PrettyTable([f for f in fields], caching=False)
    pt.align = 'l'

    for o in objs:
        row = []
        for field in fields:
            if field in formatters:
                row.append(formatters[field](o))
            else:
                if field in mixed_case_fields:
                    field_name = field.replace(' ', '_')
                else:
                    field_name = field.lower().replace(' ', '_')
                data = getattr(o, field_name, '')
                if data is None:
                    data = '-'
                # '\r' would break the table, so remove it.
                data = str(data).replace("\r", "")
                row.append(data)
        pt.add_row(row)

    if sortby is not None:
        result = encodeutils.safe_encode(pt.get_string(sortby=sortby))
    else:
        result = encodeutils.safe_encode(pt.get_string())

    result = result.decode()

    print(result)


def _flatten(data, prefix=None):
    """Flatten a dict, using name as a prefix for the keys of dict.

    >>> _flatten('cpu_info', {'arch':'x86_64'})
    [('cpu_info_arch': 'x86_64')]

    """
    if isinstance(data, dict):
        for key, value in data.items():
            new_key = '%s_%s' % (prefix, key) if prefix else key
            if isinstance(value, (dict, list)) and value:
                for item in _flatten(value, new_key):
                    yield item
            else:
                yield new_key, value
    else:
        yield prefix, data


def flatten_dict(data):
    """Return a new dict whose sub-dicts have been merged into the
    original.  Each of the parents keys are prepended to the child's
    to prevent collisions.  Any string elements will be JSON parsed
    before flattening.

    >>> flatten_dict({'service': {'host':'cloud9@compute-068', 'id': 143}})
    {'service_host': colud9@compute-068', 'service_id': 143}

    """
    data = data.copy()
    # Try and decode any nested JSON structures.
    for key, value in data.items():
        if isinstance(value, str):
            try:
                data[key] = jsonutils.loads(value)
            except ValueError:
                pass

    return dict(_flatten(data))


def print_dict(d, dict_property="Property", dict_value="Value", wrap=0):
    pt = prettytable.PrettyTable([dict_property, dict_value], caching=False)
    pt.align = 'l'
    for k, v in sorted(d.items()):
        # convert dict to str to check length
        if isinstance(v, (dict, list)):
            v = jsonutils.dumps(v, ensure_ascii=False)
        if wrap > 0:
            v = textwrap.fill(str(v), wrap)
        # if value has a newline, add in multiple rows
        # e.g. fault with stacktrace
        if v and isinstance(v, str) and (r'\n' in v or '\r' in v):
            # '\r' would break the table, so remove it.
            if '\r' in v:
                v = v.replace('\r', '')
            lines = v.strip().split(r'\n')
            col1 = k
            for line in lines:
                pt.add_row([col1, line])
                col1 = ''
        else:
            if v is None:
                v = '-'
            pt.add_row([k, v])

    result = encodeutils.safe_encode(pt.get_string())

    result = result.decode()

    print(result)


def find_resource(manager, name_or_id, wrap_exception=True, **find_args):
    """Helper for the _find_* methods."""
    # for str id which is not uuid (for Flavor, Keypair and hypervsior in cells
    # environments search currently)
    if getattr(manager, 'is_alphanum_id_allowed', False):
        try:
            return manager.get(name_or_id)
        except exceptions.NotFound:
            pass

    # first try to get entity as uuid
    try:
        tmp_id = encodeutils.safe_encode(name_or_id)

        tmp_id = tmp_id.decode()

        if uuidutils.is_uuid_like(tmp_id):
            return manager.get(tmp_id)
    except (TypeError, exceptions.NotFound):
        pass

    # then try to get entity as name
    try:
        try:
            resource = getattr(manager, 'resource_class', None)
            name_attr = resource.NAME_ATTR if resource else 'name'
            kwargs = {name_attr: name_or_id}
            kwargs.update(find_args)
            return manager.find(**kwargs)
        except exceptions.NotFound:
            pass

        # then try to find entity by human_id
        try:
            return manager.find(human_id=name_or_id, **find_args)
        except exceptions.NotFound:
            pass
    except exceptions.NoUniqueMatch:
        msg = (_("Multiple %(class)s matches found for '%(name)s', use an ID "
                 "to be more specific.") %
               {'class': manager.resource_class.__name__.lower(),
                'name': name_or_id})
        if wrap_exception:
            raise exceptions.CommandError(msg)
        raise exceptions.NoUniqueMatch(msg)

    # finally try to get entity as integer id
    try:
        return manager.get(int(name_or_id))
    except (TypeError, ValueError, exceptions.NotFound):
        msg = (_("No %(class)s with a name or ID of '%(name)s' exists.") %
               {'class': manager.resource_class.__name__.lower(),
                'name': name_or_id})
        if wrap_exception:
            raise exceptions.CommandError(msg)
        raise exceptions.NotFound(404, msg)


def format_servers_list_networks(server):
    output = []
    for (network, addresses) in server.networks.items():
        if len(addresses) == 0:
            continue
        addresses_csv = ', '.join(addresses)
        group = "%s=%s" % (network, addresses_csv)
        output.append(group)

    return '; '.join(output)


def format_security_groups(groups):
    return ', '.join(group['name'] for group in groups)


def _format_field_name(attr):
    """Format an object attribute in a human-friendly way."""
    # Split at ':' and leave the extension name as-is.
    parts = attr.rsplit(':', 1)
    name = parts[-1].replace('_', ' ')
    # Don't title() on mixed case
    if name.isupper() or name.islower():
        name = name.title()
    parts[-1] = name
    return ': '.join(parts)


def make_field_formatter(attr, filters=None):
    """
    Given an object attribute, return a formatted field name and a
    formatter suitable for passing to print_list.

    Optionally pass a dict mapping attribute names to a function. The function
    will be passed the value of the attribute and should return the string to
    display.
    """
    filter_ = None
    if filters:
        filter_ = filters.get(attr)

    def get_field(obj):
        field = getattr(obj, attr, '')
        if field and filter_:
            field = filter_(field)
        return field

    name = _format_field_name(attr)
    formatter = get_field
    return name, formatter


def safe_issubclass(*args):
    """Like issubclass, but will just return False if not a class."""

    try:
        if issubclass(*args):
            return True
    except TypeError:
        pass

    return False


def _get_resource_string(resource):
    if hasattr(resource, 'human_id') and resource.human_id:
        if hasattr(resource, 'id') and resource.id:
            return "%s (%s)" % (resource.human_id, resource.id)
        else:
            return resource.human_id
    elif hasattr(resource, 'id') and resource.id:
        return resource.id
    else:
        return resource


def do_action_on_many(action, resources, success_msg, error_msg):
    """Helper to run an action on many resources."""
    failure_flag = False

    for resource in resources:
        try:
            action(resource)
            print(success_msg % _get_resource_string(resource))
        except Exception as e:
            failure_flag = True
            print(encodeutils.safe_encode(str(e)))

    if failure_flag:
        raise exceptions.CommandError(error_msg)


def is_integer_like(val):
    """Returns validation of a value as an integer."""
    try:
        int(val)
        return True
    except (TypeError, ValueError, AttributeError):
        return False


def validate_flavor_metadata_keys(keys):
    for key in keys:
        valid_name = VALID_KEY_REGEX.match(key)
        if not valid_name:
            msg = _('Invalid key: "%s". Keys may only contain letters, '
                    'numbers, spaces, underscores, periods, colons and '
                    'hyphens.')
            raise exceptions.CommandError(msg % key)


@contextlib.contextmanager
def record_time(times, enabled, *args):
    """Record the time of a specific action.

    :param times: A list of tuples holds time data.
    :param enabled: Whether timing is enabled.
    :param args: Other data to be stored besides time data, these args
        will be joined to a string.
    """
    if not enabled:
        yield
    else:
        start = time.time()
        yield
        end = time.time()
        times.append((' '.join(args), start, end))


def prepare_query_string(params):
    """Convert dict params to query string"""
    # Transform the dict to a sequence of two-element tuples in fixed
    # order, then the encoded string will be consistent in Python 2&3.
    if not params:
        return ''
    params = sorted(params.items(), key=lambda x: x[0])
    return '?%s' % parse.urlencode(params) if params else ''


def get_url_with_filter(url, filters):
    query_string = prepare_query_string(filters)
    url = "%s%s" % (url, query_string)
    return url



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient.tests.unit.fixture_data import client
from novaclient.tests.unit.fixture_data import quotas as data
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes


class QuotaSetsTest(utils.FixturedTestCase):

    client_fixture_class = client.V1
    data_fixture_class = data.V1

    def test_tenant_quotas_get(self):
        tenant_id = 'test'
        q = self.cs.quotas.get(tenant_id)
        self.assert_request_id(q, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-quota-sets/%s' % tenant_id)
        return q

    def test_user_quotas_get(self):
        tenant_id = 'test'
        user_id = 'fake_user'
        q = self.cs.quotas.get(tenant_id, user_id=user_id)
        self.assert_request_id(q, fakes.FAKE_REQUEST_ID_LIST)
        url = '/os-quota-sets/%s?user_id=%s' % (tenant_id, user_id)
        self.assert_called('GET', url)

    def test_tenant_quotas_get_detail(self):
        tenant_id = 'test'
        q = self.cs.quotas.get(tenant_id, detail=True)
        self.assert_request_id(q, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-quota-sets/%s/detail' % tenant_id)

    def test_user_quotas_get_detail(self):
        tenant_id = 'test'
        user_id = 'fake_user'
        q = self.cs.quotas.get(tenant_id, user_id=user_id, detail=True)
        self.assert_request_id(q, fakes.FAKE_REQUEST_ID_LIST)
        url = '/os-quota-sets/%s/detail?user_id=%s' % (tenant_id, user_id)
        self.assert_called('GET', url)

    def test_tenant_quotas_defaults(self):
        tenant_id = '97f4c221bff44578b0300df4ef119353'
        q = self.cs.quotas.defaults(tenant_id)
        self.assert_request_id(q, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-quota-sets/%s/defaults' % tenant_id)

    def test_force_update_quota(self):
        q = self.cs.quotas.get('97f4c221bff44578b0300df4ef119353')
        qu = q.update(cores=2, force=True)
        self.assert_request_id(qu, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called(
            'PUT', '/os-quota-sets/97f4c221bff44578b0300df4ef119353',
            {'quota_set': {'force': True, 'cores': 2}})
        return q

    def test_quotas_delete(self):
        tenant_id = 'test'
        ret = self.cs.quotas.delete(tenant_id)
        self.assert_request_id(ret, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('DELETE', '/os-quota-sets/%s' % tenant_id)

    def test_user_quotas_delete(self):
        tenant_id = 'test'
        user_id = 'fake_user'
        ret = self.cs.quotas.delete(tenant_id, user_id=user_id)
        self.assert_request_id(ret, fakes.FAKE_REQUEST_ID_LIST)
        url = '/os-quota-sets/%s?user_id=%s' % (tenant_id, user_id)
        self.assert_called('DELETE', url)


class QuotaSetsTest2_57(QuotaSetsTest):
    """Tests the quotas API binding using the 2.57 microversion."""
    data_fixture_class = data.V2_57
    invalid_resources = ['floating_ips', 'fixed_ips', 'networks',
                         'security_groups', 'security_group_rules',
                         'injected_files', 'injected_file_content_bytes',
                         'injected_file_path_bytes']

    def setUp(self):
        super(QuotaSetsTest2_57, self).setUp()
        self.cs.api_version = api_versions.APIVersion('2.57')

    def test_tenant_quotas_get(self):
        q = super(QuotaSetsTest2_57, self).test_tenant_quotas_get()
        for invalid_resource in self.invalid_resources:
            self.assertFalse(hasattr(q, invalid_resource),
                             '%s should not be in %s' % (invalid_resource, q))

    def test_force_update_quota(self):
        q = super(QuotaSetsTest2_57, self).test_force_update_quota()
        for invalid_resource in self.invalid_resources:
            self.assertFalse(hasattr(q, invalid_resource),
                             '%s should not be in %s' % (invalid_resource, q))

    def test_update_quota_invalid_resources(self):
        """Tests trying to update quota values for invalid resources."""
        q = self.cs.quotas.get('test')
        self.assertRaises(TypeError, q.update, floating_ips=1)
        self.assertRaises(TypeError, q.update, fixed_ips=1)
        self.assertRaises(TypeError, q.update, security_groups=1)
        self.assertRaises(TypeError, q.update, security_group_rules=1)
        self.assertRaises(TypeError, q.update, networks=1)
        self.assertRaises(TypeError, q.update, injected_files=1)
        self.assertRaises(TypeError, q.update, injected_file_content_bytes=1)
        self.assertRaises(TypeError, q.update, injected_file_path_bytes=1)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from novaclient.tests.functional import base


class TestMigrationList(base.ClientTestBase):
    """Tests the "nova migration-list" command."""

    def _filter_migrations(
            self, version, migration_type, source_compute):
        """
        Filters migrations by --migration-type and --source-compute.

        :param version: The --os-compute-api-version to use.
        :param migration_type: The type of migrations to filter.
        :param source_compute: The source compute service hostname to filter.
        :return: output of the nova migration-list command with filters applied
        """
        return self.nova('migration-list',
                         flags='--os-compute-api-version %s' % version,
                         params='--migration-type %s --source-compute %s' % (
                             migration_type, source_compute))

    def test_migration_list(self):
        """Tests creating a server, resizing it and then listing and filtering
        migrations using various microversion milestones.
        """
        server_id = self._create_server(flavor=self.flavor.id).id
        # Find the source compute by getting OS-EXT-SRV-ATTR:host from the
        # nova show output.
        server = self.nova('show', params='%s' % server_id)
        server_user_id = self._get_value_from_the_table(server, 'user_id')
        tenant_id = self._get_value_from_the_table(server, 'tenant_id')
        source_compute = self._get_value_from_the_table(
            server, 'OS-EXT-SRV-ATTR:host')
        # now resize up
        alternate_flavor = self._pick_alternate_flavor()
        self.nova('resize',
                  params='%s %s --poll' % (server_id, alternate_flavor))
        # now confirm the resize
        self.nova('resize-confirm', params='%s' % server_id)
        # wait for the server to be active and then check the migration list
        self._wait_for_state_change(server_id, 'active')
        # First, list migrations with v2.1 and our server id should be in the
        # output. There should only be the one migration.
        migrations = self.nova('migration-list',
                               flags='--os-compute-api-version 2.1')
        instance_uuid = self._get_column_value_from_single_row_table(
            migrations, 'Instance UUID')
        self.assertEqual(server_id, instance_uuid)
        # A successfully confirmed resize should have the migration status
        # of "confirmed".
        migration_status = self._get_column_value_from_single_row_table(
            migrations, 'Status')
        self.assertEqual('confirmed', migration_status)
        # Now listing migrations with 2.23 should give us the Type column which
        # should have a value of "resize".
        migrations = self.nova('migration-list',
                               flags='--os-compute-api-version 2.23')
        migration_type = self._get_column_value_from_single_row_table(
            migrations, 'Type')
        self.assertEqual('resize', migration_type)
        # Filter migrations with v2.1.
        migrations = self._filter_migrations('2.1', 'resize', source_compute)
        # Make sure we got something back.
        src_compute = self._get_column_value_from_single_row_table(
            migrations, 'Source Compute')
        self.assertEqual(source_compute, src_compute)
        # Filter migrations with v2.59 and make sure there is a migration UUID
        # value in the output.
        migrations = self._filter_migrations('2.59', 'resize', source_compute)
        # _get_column_value_from_single_row_table will raise ValueError if a
        # value is not found for the given column. We don't actually care what
        # the migration UUID value is just that the filter works and the UUID
        # is shown.
        self._get_column_value_from_single_row_table(migrations, 'UUID')
        # Filter migrations with v2.66, same as 2.59.
        migrations = self._filter_migrations('2.66', 'resize', source_compute)
        self._get_column_value_from_single_row_table(migrations, 'UUID')
        # Now do a negative test to show that filtering on a migration type
        # that we don't have a migration for will not return anything.
        migrations = self._filter_migrations(
            '2.1', 'evacuation', source_compute)
        self.assertNotIn(server_id, migrations)
        # Similarly, make sure we don't get anything back when filtering on
        # a --source-compute that doesn't exist.
        migrations = self._filter_migrations(
            '2.66', 'resize', uuidutils.generate_uuid())
        self.assertNotIn(server_id, migrations)

        # Listing migrations with v2.80 and make sure there are the User ID
        # and Project ID values in the output.
        migrations = self.nova('migration-list',
                               flags='--os-compute-api-version 2.80')
        user_id = self._get_column_value_from_single_row_table(
            migrations, 'User ID')
        self.assertEqual(server_user_id, user_id)
        project_id = self._get_column_value_from_single_row_table(
            migrations, 'Project ID')
        self.assertEqual(tenant_id, project_id)
